# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

from typing import List
import pymonetdb
import unittest
from tests.util import test_args


class ConnectionTest(unittest.TestCase):

    connections_to_close: List[pymonetdb.Connection] = []
    schemas_to_drop: List[str] = []

    def tearDown(self):
        with self.connect() as conn, conn.cursor() as c:
            for s in self.schemas_to_drop:
                try:
                    c.execute("DROP SCHEMA %s", (s,))
                except pymonetdb.OperationalError:
                    pass
        for conn in self.connections_to_close:
            try:
                conn.close()
            except pymonetdb.exceptions.Error:
                pass
        self.connections_to_close = []
        super().tearDown()

    def connect(self, **more_args) -> pymonetdb.Connection:
        args = {**test_args, **more_args}
        conn = pymonetdb.connect(**args)
        self.connections_to_close.append(conn)
        return conn

    def test_autocommit_off(self):
        conn = self.connect(autocommit=False)
        # Verify that the connection works at all
        with conn.cursor() as c:
            c.execute("SELECT 42")
            # In autocommit mode, rollbacks fail
            conn.rollback()

    def test_autocommit_on(self):
        conn = self.connect(autocommit=True)
        # Verify that the connection works at all
        with conn.cursor() as c:
            c.execute("SELECT 42")
            # In autocommit mode, rollbacks fail
            with self.assertRaisesRegex(pymonetdb.OperationalError, "auto commit"):
                conn.rollback()

    def test_timezone(self):
        minutes_east = 60 + 23
        conn = self.connect(timezone=minutes_east)
        with conn.cursor() as c:
            c.execute("SELECT now()")
            now = c.fetchone()[0]
            tz = now.tzname()
            self.assertIn("+01:23", tz)

    def test_replysize(self):
        conn = self.connect(replysize=99)
        with conn.cursor() as c:
            self.assertEqual(c.replysize, 99)

    def test_schema(self):
        conn = self.connect()
        with conn.cursor() as c:
            c.execute("SELECT current_schema")
            cur = c.fetchone()[0]
            c.execute("SELECT name FROM sys.schemas")
            existing_schemas = [row[0] for row in c.fetchall()]
            for s in existing_schemas:
                schema_to_connect_to = s
                if schema_to_connect_to not in [cur, 'tmp']:
                    break
            else:
                self.fail(f"no suitable schema in {existing_schemas!r}")

        # no try to connect with this schema set
        conn = self.connect(schema=schema_to_connect_to)
        with conn.cursor() as c:
            c.execute("SELECT current_schema")
            actual_schema = c.fetchone()[0]
            self.assertEqual(actual_schema, schema_to_connect_to)



#!/usr/bin/env python3

import logging
logging.basicConfig(level=logging.DEBUG)

import pymonetdb

#conn = pymonetdb.connect('demo', port=44001, replysize=1, binary=False)
conn = pymonetdb.connect('monetdb://127.0.0.1:50000/demo')
c = conn.cursor()

c.execute("SELECT value from sys.generate_series(0,9); SELECT 100 + value as i, 1000 + value as j FROM sys.generate_series(0,100_000)")

go = True
while go:
    print('RESULT SET', 50 * '-')
    for row in c:
        print(row)
    go = c.nextset()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.


import hashlib
from ssl import SSLError
from tempfile import NamedTemporaryFile
from typing import Any, Dict, Optional, Union
from unittest import SkipTest, TestCase, skipUnless
from urllib.parse import quote as urlquote
import urllib.request

import pymonetdb
from pymonetdb.exceptions import DatabaseError

from tests.util import (
    test_tls_tester_host,
    test_tls_tester_port,
    test_tls_tester_sys_store,
)

import logging
logging.basicConfig(level=logging.DEBUG)


class TestTLS(TestCase):
    _name: Optional[str]
    _cache: Dict[str, str]
    _files: Dict[str, Any]  # they are NamedTemporaryFile's, but mypy hates those

    def __init__(self, methodName):
        self._name = methodName
        self._cache = dict()
        self._files = dict()
        super().__init__(methodName)

    def setUp(self) -> None:
        # validate the config
        have_host = test_tls_tester_host is not None
        have_port = test_tls_tester_port is not None
        have_sys = test_tls_tester_sys_store
        if have_host != have_port:
            raise Exception(
                "Either pass both TSTTLSTESTERHOST and TSTTLSTESTERPORT, or neither"
            )
        if have_sys and not have_host:
            raise Exception(
                "Setting TSTTLSTESTERSYSSTORE does not make sense without TSTTLSTESTERHOST and TSTTLSTESTERPORT"
            )

        if not have_host:
            raise SkipTest("TSTTLSTESTERHOST and TSTTLSTESTERPORT not set")

    def try_connect(self, port_name: str, tls=True, cert=None, expect=None, **kwargs):
        """Try to connect to the named port, looking it up in tlstester.py's portmap.

        Returns succesfully if pymonetdb.connect raised a DatabaseError containing
        tlstester's signature message. This indicates that a MAPI dialogue took
        place. Otherwise, raise whatever exception raised by pymonetdb.connect.
        """

        port = self.port(port_name)

        try:
            pymonetdb.connect(
                "banana",
                hostname=test_tls_tester_host,
                port=port,
                tls=tls,
                cert=cert,
                **kwargs,
            )
            self.fail("Expected connection to tlstester.py to fail but it didn't")
        except DatabaseError as e:
            msg = str(e)
            if "Sorry, this is not a real MonetDB instance" not in msg:
                raise
            if expect is None:
                expect = port_name
            if expect and f"({expect})" not in msg:
                raise

    def port(self, port_name: str) -> int:
        portmap = dict()
        url = f"/?test={urlquote(self._name)}" if self._name else "/"
        ports = self.download(url, encoding="utf-8")
        assert isinstance(ports, str)   # silence mypy
        for line in ports.splitlines():
            name_field, port_field = line.split(":", 1)
            portmap[name_field] = int(port_field)
        port = portmap.get(port_name)
        if port is None:
            names = ", ".join(repr(n) for n in portmap.keys())
            raise Exception(
                f"tlstester.py didn't offer port '{port_name}', only {names}"
            )
        return port

    def download(self, path, encoding=None) -> Union[bytes, str]:
        if path in self._cache:
            content = self._cache[path]
        else:
            url = f"http://{test_tls_tester_host}:{test_tls_tester_port}{path}"
            with urllib.request.urlopen(url) as resp:
                content = resp.read()
            self._cache[path] = content

        assert isinstance(content, bytes)
        if encoding:
            return str(content, encoding=encoding)
        else:
            return content

    def download_file(self, path: str) -> str:
        if path in self._files:
            file = self._files[path]
        else:
            content = self.download(path, encoding=None)
            file = NamedTemporaryFile()
            file.write(content)
            file.flush()
            self._files[path] = file
        return file.name

    def test_connect_plain(self):
        self.try_connect("plain", tls=False)

    def test_connect_tls(self):
        self.try_connect("server1", cert=self.download_file("/ca1.crt"))

    def test_refuse_no_cert(self):
        with self.assertRaisesRegex(
            SSLError, "self[- ]signed certificate in certificate chain"
        ):
            self.try_connect("server1", cert=None)

    def test_refuse_wrong_cert(self):
        with self.assertRaisesRegex(
            SSLError, "self[- ]signed certificate in certificate chain"
        ):
            self.try_connect("server1", cert=self.download_file("/ca2.crt"))

    def test_refuse_tls12(self):
        with self.assertRaisesRegex(SSLError, "version"):
            self.try_connect("tls12", cert=self.download_file("/ca1.crt"))

    def test_refuse_expired(self):
        with self.assertRaisesRegex(SSLError, "has expired"):
            self.try_connect("expiredcert", cert=self.download_file("/ca1.crt"))

    def test_connect_client_auth(self):
        self.try_connect(
            "clientauth",
            cert=self.download_file("/ca1.crt"),
            clientkey=self.download_file("/client2.key"),
            clientcert=self.download_file("/client2.crt"),
        )

    def test_connect_client_auth_combined(self):
        self.try_connect(
            "clientauth",
            cert=self.download_file("/ca1.crt"),
            clientkey=self.download_file("/client2.keycrt"),
        )

    def test_fail_plain_to_tls(self):
        with self.assertRaises(SSLError):
            self.try_connect("plain")

    def test_fail_tls_to_plain(self):
        with self.assertRaises(ConnectionError):
            self.try_connect("server1", tls=False)

    def get_fingerprint(self, name: str, prefix=True) -> str:
        algo = 'sha256'
        ndigits = 6
        file_name = self.download_file(name)
        with open(file_name, "rb") as f:
            der = f.read()
        digest = hashlib.new(algo, der).hexdigest()
        fingerprint = digest[:ndigits]
        if prefix:
            fingerprint = algo + ":" + fingerprint
        return fingerprint

    def test_connect_certhash(self):
        self.try_connect("server1", certhash=self.get_fingerprint("/server1.der"))

    def test_fail_connect_wrong_fingerprint(self):
        # we connect to server1 but use the fingerprint of server2's cert
        with self.assertRaisesRegex(SSLError, "wrong server certificate hash"):
            self.try_connect("server1", certhash=self.get_fingerprint("/server2.der"))

    def test_connect_redirected(self):
        self.try_connect("redirect", expect="server2", cert=self.download_file("/ca1.crt"))

    @skipUnless(test_tls_tester_sys_store, "TSTTLSTESTERSYSSTORE not set")
    def test_connect_trusted(self):
        self.try_connect("server3", cert=None)



from unittest import TestCase
from pymonetdb import connect
from tests.util import test_args


class TestPrepare(TestCase):
    def setUp(self):
        self.connection = connect(autocommit=False, **test_args)
        self.cursor = self.connection.cursor()

    def test_prepare(self):
        # It would be nice to have a prepare API but in the mean time we can
        # invoke PREPARE and EXEC ourselves.
        #
        # Note how the id of the newly prepared statement is stashed in
        # cursor.lastrowid.

        prepare = "PREPARE SELECT value, 10 * value, 100 * value FROM sys.generate_series(0, ?)"
        self.cursor.execute(prepare)
        exec_id = self.cursor.lastrowid

        n = 3
        expected = [
            (value, 10 * value, 100 * value)
            for value in range(n)
        ]
        self.cursor.execute("EXEC %s(%s)", (exec_id, n))
        received = self.cursor.fetchall()
        self.assertEqual(expected, received)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

import unittest

import pymonetdb

# regular import doesn't work, don't know why
DATE = pymonetdb.types.DATE
TIME = pymonetdb.types.TIME
TIMESTAMP = pymonetdb.types.TIMESTAMP
TIMETZ = pymonetdb.types.TIMETZ
TIMESTAMPTZ = pymonetdb.types.TIMESTAMPTZ


class TestPythonizeTemporal(unittest.TestCase):

    def verify(self, val, typ, expected=None):
        dt = pymonetdb.sql.pythonize.convert(val, typ)
        iso = dt.isoformat()
        if expected is None:
            expected = val.replace(' ', 'T')
        self.assertEqual(iso, expected)

    def test_date(self):
        self.verify('2015-02-14', DATE)
        self.verify('1970-02-14', DATE)
        self.verify('1950-02-14', DATE)
        self.verify('1676-02-14', DATE)
        self.verify('2-02-14', DATE, '0002-02-14')

    def test_time(self):
        self.verify('00:00:00', TIME)
        self.verify('00:00:00.00', TIME, '00:00:00')
        self.verify('00:00:00.12', TIME, '00:00:00.120000')
        self.verify('12:13:14.1516', TIME, '12:13:14.151600')
        self.verify('23:59:59.999999', TIME)
        with self.assertRaises(ValueError):
            # hello, leap second
            self.verify('23:59:60.123456', TIME)

    def test_timetz(self):
        self.verify('00:00:00+01:30', TIMETZ, '00:00:00+01:30')
        self.verify('12:34:56-04:00', TIMETZ, '12:34:56-04:00')

    def test_timestamp(self):
        self.verify('2015-02-14 20:50:12', TIMESTAMP)
        self.verify('2015-02-14 20:50:12.34', TIMESTAMP, '2015-02-14T20:50:12.340000')
        self.verify('15-02-14 20:50:12.34', TIMESTAMP, '0015-02-14T20:50:12.340000')

    def test_timestamptz(self):
        self.verify('2015-02-14 20:50:12-04:30', TIMESTAMPTZ)
        self.verify('2015-02-14 20:50:12.34+04:30', TIMESTAMPTZ, '2015-02-14T20:50:12.340000+04:30')
        self.verify('15-02-14 20:50:12.34-04:30', TIMESTAMPTZ, '0015-02-14T20:50:12.340000-04:30')



"""
This is the python implementation of the mapi protocol.
"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.


import os
import re
import socket
import logging
import struct
import hashlib
import ssl
import typing
from typing import Callable, List, Optional, Tuple, Union

from pymonetdb.exceptions import OperationalError, DatabaseError, \
    ProgrammingError, NotSupportedError, IntegrityError
from pymonetdb.target import Target

if typing.TYPE_CHECKING:
    from pymonetdb.filetransfer.downloads import Downloader
    from pymonetdb.filetransfer.uploads import Uploader

logger = logging.getLogger(__name__)

MAX_PACKAGE_LENGTH = (1024 * 8) - 2

MSG_PROMPT = ""
MSG_MORE = "\1\2\n"
MSG_FILETRANS = "\1\3\n"
MSG_INFO = "#"
MSG_ERROR = "!"
MSG_Q = "&"
MSG_QTABLE = "&1"
MSG_QUPDATE = "&2"
MSG_QSCHEMA = "&3"
MSG_QTRANS = "&4"
MSG_QPREPARE = "&5"
MSG_QBLOCK = "&6"
MSG_HEADER = "%"
MSG_TUPLE = "["
MSG_TUPLE_NOSLICE = "="
MSG_REDIRECT = "^"
MSG_OK = "=OK"

MSG_ERROR_B = bytes(MSG_ERROR, 'ascii')
MSG_FILETRANS_B = bytes(MSG_FILETRANS, 'ascii')

STATE_INIT = 0
STATE_READY = 1

# MonetDB error codes
errors = {
    '42S02': OperationalError,  # no such table
    '40002': IntegrityError,  # INSERT INTO: UNIQUE constraint violated
    '2D000': IntegrityError,  # COMMIT: failed
    '40000': IntegrityError,  # DROP TABLE: FOREIGN KEY constraint violated
    'M0M29': IntegrityError,  # The code monetdb emitted before Jun2020
}


def handle_error(error):
    """Return exception matching error code.

    args:
        error (str): error string, potentially containing mapi error code

    returns:
        tuple (Exception, formatted error): returns OperationalError if unknown
            error or no error code in string

    """

    if error[:13] == 'SQLException:':
        idx = str.index(error, ':', 14)
        error = error[idx + 10:]
    if len(error) > 5 and error[:5] in errors:
        return errors[error[:5]], error
    else:
        return OperationalError, error


# noinspection PyExceptionInherit
class Connection(object):
    """
    MAPI (low level MonetDB API) connection
    """

    state: int = STATE_INIT
    target: Target = Target()
    socket: Optional[Union['socket.socket', ssl.SSLSocket]] = None
    is_tcp: Optional[bool] = None
    is_raw_control: Optional[bool] = None
    handshake_options_callback: Optional[Callable[[int], List['HandshakeOption']]] = None
    remaining_handshake_options: List['HandshakeOption'] = []
    uploader: Optional['Uploader'] = None
    downloader: Optional['Downloader'] = None
    stashed_buffer: Optional[bytearray] = None

    def connect(self, database: Optional[Union[Target, str]] = None, *args, **kwargs):  # noqa C901
        """ setup connection to MAPI server
        """

        # Ideally we'd just take the Target as a parameter, but we want to
        # provide some backward compatibility so we first deal with the legacy
        # arguments.
        callback = kwargs.get('handshake_options_callback')
        if callback is not None:
            self.handshake_options_callback = callback
            del kwargs['handshake_options_callback']

        # Create Target or use given
        if isinstance(database, Target):
            self.target = database.clone()
            assert not args and not kwargs
        else:
            self.target = construct_target_from_args(database, *args, **kwargs)

        # Validate the target parameters
        try:
            self.target.validate()
        except ValueError as e:
            raise DatabaseError(str(e))

        if self.target.connect_scan:
            self.scan_sockdir()
            return

        # Close any remainders of previous attempts
        if self.socket:
            try:
                self.socket.close()
            except OSError:
                pass
            self.socket = None

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Connecting to {self.target.summary_url()}")
        # Enter a loop to deal with redirects.
        try:
            self.connect_loop()
        except Exception as e:
            logger.error(f"Could not connect to {self.target.summary_url()}: {e}")
            raise
        if logger.isEnabledFor(logging.INFO):
            logger.info(f"Established connection to {self.target.summary_url()}")

        # We have a working connection now. Take care of the options we couldn't
        # handle during the handshake
        self.state = STATE_READY

        for opt in self.remaining_handshake_options:
            opt.fallback(opt.value)

    def connect_loop(self):
        for i in range(10):
            # maybe the previous attempt left an open socket that just needs an
            # additional login attempt
            if self.socket is None:
                # No, we need to make a new connection
                self.try_connect()
                assert self.socket is not None

                if self.target.connect_timeout:
                    # The new socket's timeout was overridden during the
                    # connect. Put it back.
                    self.socket.settimeout(socket.getdefaulttimeout())

                # Once connected, deal with the file handle passing protocol,
                # AND with TLS. Note that these are necessarily exclusive, we
                # can't do TLS over unix domain sockets.
                self.is_raw_control = False
                if self.is_tcp:
                    self.prime_or_wrap_connection()
                elif self.target.language == 'control':
                    self.is_raw_control = True
                else:
                    # Send a '0' (0x48) to let the other side know we're not
                    # going to try to pass a file handle.
                    self.socket.sendall(b'0')

            # We have a connection now. Try to log in. If it succeeds, we're
            # done. If it fails, _login should either
            # 1) close the socket and update self.target for a new attempt, or
            # 2) leave the socket open for another login attempt.
            if self.is_raw_control:
                # no login needed, we're done
                break
            elif self._login():
                break
            else:
                # _login has determined that we need another round
                continue
        else:
            raise OperationalError("too many redirects")

    def try_connect(self):  # noqa C901
        err = None
        timeout = self.target.connect_timeout

        sock = self.target.connect_unix
        if sock and hasattr(socket, 'AF_UNIX'):
            s = socket.socket(socket.AF_UNIX)
            if timeout:
                s.settimeout(float(timeout))
            try:
                s.connect(sock)
                # it worked!
                logger.debug(f"Connected to {sock}")
                self.socket = s
                self.is_tcp = False
                return
            except OSError as e:
                s.close()
                err = e

        host = self.target.connect_tcp
        if host:
            port = self.target.connect_port
            addrs = socket.getaddrinfo(host, port, socket.AF_UNSPEC, socket.SOCK_STREAM)
            for fam, typ, proto, cname, addr in addrs:
                s = socket.socket(fam, typ, proto)
                s.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
                s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                if timeout:
                    s.settimeout(float(timeout))
                try:
                    s.connect(addr)
                    # it worked!
                    logger.debug(f"Connected to {addr[0]} port {addr[1]}")
                    self.socket = s
                    self.is_tcp = True
                    return
                except OSError as e:
                    s.close()
                    err = e

        if err is not None:
            raise err
        raise DatabaseError("endpoint not found")

    def prime_or_wrap_connection(self):  # noqa: C901
        if not self.target.tls:
            # Prime the connection with some NUL bytes.
            # We expect the remote server to be a MAPI server, in which
            # case it will ignore them.
            # But if it is accidentally a TLS server, the NUL bytes tend
            # to force an error, avoiding a hang.
            # Also, unexpectedly, in some situations sending the NUL bytes
            # appear to make connection setup a little faster rather than slower.
            self.socket.sendall(b'\x00\x00\x00\x00\x00\x00\x00\x00')
            return

        target = self.target
        verification_mode = target.connect_tls_verify

        if target.dangerous_tls_nocheck:
            disabled_checks = set(target.dangerous_tls_nocheck.split(','))
        else:
            disabled_checks = set()

        # Set up the SSL context. How to do that depends on the verification mode
        if verification_mode == 'system':
            ssl_context = ssl.create_default_context()
        elif verification_mode == 'cert':
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            ssl_context.load_verify_locations(target.cert)
        else:
            assert verification_mode == 'hash'
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            disabled_checks.add('host')
            disabled_checks.add('cert')

        # The following is common between all verification modes
        ssl_context.minimum_version = ssl.TLSVersion.TLSv1_3
        ssl_context.set_alpn_protocols(["mapi/9"])
        if target.clientkey:
            certfile = (target.clientcert or target.clientkey)
            keyfile = target.clientkey
            ssl_context.load_cert_chain(certfile, keyfile)
        if 'host' in disabled_checks:
            ssl_context.check_hostname = False
        if 'cert' in disabled_checks:
            ssl_context.verify_mode = ssl.CERT_NONE

        # Perform the SSL handshake and switch to the encrypted connection
        self.socket = ssl_context.wrap_socket(self.socket, server_hostname=target.connect_tcp)

        # In hash mode, verify the identity of the server. Otherwise, just log a message about it
        if verification_mode == 'hash':
            self._verify_fingerprint(target.certhash)
            logger.debug(f"TLS certificate matches hash {target.certhash}")
        else:
            peercert = self.socket.getpeercert()
            if peercert:
                logger.debug("Valid TLS certificate")
            else:
                logger.debug("TLS certificate check was disabled")

    def _login(self) -> bool:  # noqa: C901
        """ Reads challenge from line, generate response and check if
        everything is okay """

        assert self.socket

        challenge = self._getblock()
        response = self._challenge_response(challenge)
        self._putblock(response)
        prompt = self._getblock().strip()

        if len(prompt) == 0 or prompt == MSG_OK:
            # server is happy
            return True
        elif prompt.startswith(MSG_INFO):
            # is this right?
            logger.info("%s" % prompt[1:])
            return True
        elif prompt.startswith(MSG_ERROR):
            logger.error(prompt[1:])
            raise DatabaseError(prompt[1:])
        elif prompt.startswith(MSG_REDIRECT):
            # a redirect can contain multiple redirects, we only use the first
            redirect = prompt.split('\n', 1)[0][1:]
            self._handle_redirect(redirect)
            return False
        else:
            raise ProgrammingError("unknown state: %s" % prompt)
        assert False and "unreachable"

    def _handle_redirect(self, redirect: str):
        if redirect.startswith('mapi:merovingian:'):
            if not redirect.startswith('mapi:merovingian://proxy'):
                raise DatabaseError(f"Invalid redirect: {redirect}")
            logger.debug("Local redirect, restarting authentication")
        else:
            logger.debug("Redirected to " + redirect)
            try:
                self.target.parse(redirect)
            except ValueError as e:
                raise DatabaseError(str(e))
            # close the socket so the next iteration will reconnect based on the
            # updated target.
            if self.socket:
                self.socket.close()
                self.socket = None

    def _verify_fingerprint(self, fingerprint: str):
        assert self.socket and isinstance(self.socket, ssl.SSLSocket)

        m = re.match(r'sha256:([0-9a-fA-F:]+)$', fingerprint)
        if not m:
            raise ssl.SSLError(f"invalid certificate hash {fingerprint!r}")
        digits = m.group(1).lower().replace(':', '')

        der = self.socket.getpeercert(binary_form=True)
        if not der:
            raise ssl.SSLError("server has no certificate")

        digest = hashlib.sha256(der).hexdigest()
        if not digest.startswith(digits):
            raise ssl.SSLError(f"wrong server certificate hash: {fingerprint}")

    def scan_sockdir(self):   # noqa C901
        try:
            my_uid = os.getuid()
        except AttributeError:
            # Windows
            my_uid = -1

        # Scan the sockdir and put candidate sockets in
        # my_socks if they are owned by me and strange_socks otherwise
        my_socks = []
        strange_socks = []
        prefix = self.target.sockprefix
        try:
            logger.debug(f"scanning {self.target.sockdir!r} for Unix domain sockets")
            entries = [e for e in os.scandir(self.target.sockdir) if e.name.startswith(prefix)]
        except OSError:
            entries = []
        for entry in entries:
            try:
                portno = int(entry.name[len(prefix):])
                if portno < 1 or portno > 65535:
                    continue
            except ValueError:
                continue

            try:
                st = entry.stat()
            except OSError:
                continue
            if st.st_uid == my_uid:
                my_socks.append(entry.path)
            else:
                strange_socks.append(entry.path)

        # Try to connect to each of them
        for sock in my_socks + strange_socks:

            self.target.sock = sock
            try:
                logger.debug(f"Trying {sock!r}")
                self.connect(self.target)
                # If it works, use this
                return
            except Exception:
                pass
        self.target.sock = ''

        # last resort
        logger.debug("Trying a TCP connection to localhost")
        self.target.host = 'localhost'
        self.connect(self.target)

    def disconnect(self):
        """ disconnect from the monetdb server """
        logger.info("Closing connection")
        self.state = STATE_INIT
        self.socket.close()

    def _sabotage(self):
        """ Kill the connection in a way that the server is sure to recognize as an error"""
        sock = self.socket
        self.socket = None
        self.state = STATE_INIT
        if not sock:
            return
        bad_header = struct.pack('<H', 2 * 8193 + 0)  # larger than allowed, and not the final message
        bad_body = b"ERROR\x80ERROR"  # invalid utf-8, and too small
        try:
            sock.send(bad_header + bad_body)
            # and then we hang up
            sock.close()
        except Exception:
            # don't care
            pass

    def cmd(self, operation: str):  # noqa: C901
        """ put a mapi command on the line"""
        logger.debug("executing command %s" % operation)

        if self.state != STATE_READY:
            raise ProgrammingError("Not connected")

        self._putblock(operation)
        response = self._getblock_and_transfer_files()
        if not len(response):
            return ""
        elif response.startswith(MSG_OK):
            return response[3:].strip() or ""
        if response == MSG_MORE:
            # tell server it isn't going to get more
            return self.cmd("")

        # If we are performing an update test for errors such as a failed
        # transaction.

        # We are splitting the response into lines and checking each one if it
        # starts with MSG_ERROR. If this is the case, find which line records
        # the error and use it to call handle_error.
        if response[:2] == MSG_QUPDATE:
            lines = response.split('\n')
            if any([line.startswith(MSG_ERROR) for line in lines]):
                index = next(i for i, v in enumerate(lines) if v.startswith(MSG_ERROR))
                exception, msg = handle_error(lines[index][1:])
                raise exception(msg)

        if response[0] in [MSG_Q, MSG_HEADER, MSG_TUPLE]:
            return response
        elif response[0] == MSG_ERROR:
            exception, msg = handle_error(response[1:])
            raise exception(msg)
        elif response[0] == MSG_INFO:
            logger.info("%s" % (response[1:]))
        elif self.is_raw_control:
            if response.startswith("OK"):
                return response[2:].strip() or ""
            else:
                return response
        else:
            raise ProgrammingError("unknown state: %s" % response)

    def binary_cmd(self, operation: str) -> memoryview:
        """ put a mapi command on the line, with a binary response.

        returns a memoryview that can only be used until the next
        operation on this Connection object.
        """
        logger.debug("executing binary command %s" % operation)

        if self.state != STATE_READY:
            raise ProgrammingError("Not connected")

        self._putblock(operation)
        buffer = self._get_buffer()
        n = self._getblock_raw(buffer, 0)
        view = memoryview(buffer)[:n]
        self._stash_buffer(buffer)

        # Handle !Error message
        if view[0:len(MSG_ERROR_B)] == MSG_ERROR_B:
            msg_bytes = bytes(view)
            idx = msg_bytes.find(b'\n')
            if idx > 0:
                msg_bytes = msg_bytes[1:idx + 1]
            exception, msg = handle_error(str(msg_bytes, 'utf-8'))
            raise exception(msg)

        return view

    def _challenge_response(self, challenge: str):  # noqa: C901
        """ generate a response to a mapi login challenge """

        challenges = challenge.split(':')
        if challenges[-1] != '' or len(challenges) < 7:
            raise OperationalError("Server sent invalid challenge")
        challenges.pop()

        salt, server_type, protocol, hashes, endian = challenges[:5]

        if server_type == 'merovingian' and self.target.language != 'control':
            # we want to be forwarded, hide real credentials
            user = 'merovingian'
            password = ''
        else:
            user = self.target.user or ''
            password = self.target.password or ''

        if endian == 'LIT':
            self.server_endian = 'little'
        elif endian == 'BIG':
            self.server_endian = 'big'
        else:
            raise NotSupportedError('Unknown byte order: ' + endian)

        if protocol == '9':
            algo = challenges[5]
            try:
                h = hashlib.new(algo)
                h.update(password.encode())
                password = h.hexdigest()
            except ValueError as e:
                raise NotSupportedError(str(e))
        else:
            raise NotSupportedError("We only speak protocol v9")

        for i in hashes.split(","):
            try:
                s = hashlib.new(i)
            except ValueError:
                pass
            else:
                s.update(password.encode())
                s.update(salt.encode())
                pwhash = "{" + i + "}" + s.hexdigest()
                break
        else:
            raise NotSupportedError("Unsupported hash algorithms required"
                                    " for login: %s" % hashes)

        response = ":".join([
            "BIG",
            user,
            pwhash,
            self.target.language,
            self.target.database or ''
        ]) + ":"

        self.binexport_level = 0
        if len(challenges) >= 8:
            part = challenges[7]
            assert part.startswith('BINARY=')
            self.binexport_level = int(part[7:])

        callback = self.handshake_options_callback
        handshake_options = callback(self.binexport_level) if callback else []

        if len(challenges) >= 7:
            response += "FILETRANS:"
            options_level = 0
            for part in challenges[6].split(","):
                if part.startswith("sql="):
                    try:
                        options_level = int(part[4:])
                    except ValueError:
                        raise OperationalError("invalid sql options level in server challenge: " + part)
            options = []
            for opt in handshake_options:
                if opt.level < options_level:
                    options.append(opt.name + "=" + str(int(opt.value)))
                    opt.sent = True
            response += ",".join(options) + ":"

        self.remaining_handshake_options = [opt for opt in handshake_options if not opt.sent]

        return response

    def _getblock_and_transfer_files(self) -> str:
        """ read one mapi encoded block and take care of any file transfers the server requests"""
        if self.is_raw_control:
            # control connections do not use the blocking protocol and do not transfer files
            return self._recv_to_end()

        buffer = self._get_buffer()
        offset = 0

        # import this here to solve circular import
        from pymonetdb.filetransfer import handle_file_transfer

        while True:
            old_offset = offset
            offset = self._getblock_raw(buffer, old_offset)
            i = buffer.rfind(b'\n', old_offset, offset - 1)
            if i >= old_offset + 2 and buffer[i - 2: i + 1] == MSG_FILETRANS_B:
                # File transfer request. Chop the cmd off the buffer by lowering the offset
                cmd = str(buffer[i + 1: offset - 1], 'utf-8')
                offset = i - 2
                handle_file_transfer(self, cmd)
                continue
            else:
                break
        self._stash_buffer(buffer)
        return str(memoryview(buffer)[:offset], 'utf-8')

    def _getblock(self) -> str:
        """ read one mapi encoded block """
        if self.is_raw_control:
            # control connections do not use the blocking protocol
            return self._recv_to_end()
        buf = self._get_buffer()
        end = self._getblock_raw(buf, 0)
        ret = str(memoryview(buf)[:end], 'utf-8')
        self._stash_buffer(buf)
        return ret

    def _getblock_raw(self, buffer: bytearray, offset: int) -> int:
        """
        Read one mapi block into 'buffer' starting at 'offset', enlarging the buffer
        as necessary and returning offset plus the number of bytes read.
        """
        last = False
        while not last:
            offset, last = self._get_minor_block(buffer, offset)
        return offset

    def _get_minor_block(self, buffer: bytearray, offset: int) -> Tuple[int, bool]:
        self._getbytes(buffer, offset, 2)
        unpacked = buffer[offset] + 256 * buffer[offset + 1]
        length = unpacked >> 1
        last = unpacked & 1
        if length:
            offset = self._getbytes(buffer, offset, length)
        return (offset, bool(last))

    def _getbytes(self, buffer: bytearray, offset: int, count: int) -> int:
        """
        Read 'count' bytes from the socket into 'buffer' starting at 'offset'.
        Enlarge buffer if necessary.
        Return offset + count if all goes well.
        """
        assert self.socket
        end = count + offset
        if len(buffer) < end:
            # enlarge
            nblocks = 1 + (end - len(buffer)) // 8192
            buffer += bytes(nblocks * 8192)
        while offset < end:
            view = memoryview(buffer)[offset:end]
            n = self.socket.recv_into(view)
            if n == 0:
                raise BrokenPipeError("Server closed connection")
            offset += n
        return end

    def _recv_to_end(self) -> str:
        """
        Read bytes from the socket until the server closes the connection
        """
        parts = []
        while True:
            assert self.socket
            received = self.socket.recv(4096)
            if not received:
                break
            parts.append(received)
        return str(b"".join(parts).strip(), 'utf-8')

    def _get_buffer(self) -> bytearray:
        """Retrieve a previously stashed buffer for reuse, or create a new one"""
        if self.stashed_buffer:
            buffer = self.stashed_buffer
            self.stashed_buffer = None
        else:
            buffer = bytearray(8192)
        return buffer

    def _stash_buffer(self, buffer):
        """Stash a used buffer for future reuse"""
        if self.stashed_buffer is None or len(self.stashed_buffer) < len(buffer):
            self.stashed_buffer = buffer

    def _putblock(self, block):
        """ wrap the line in mapi format and put it into the socket """
        data = block.encode('utf-8')
        if self.is_raw_control:
            # control does not use the blocking protocol
            return self._send_all_and_shutdown(data)
        else:
            self._putblock_raw(block.encode(), True)

    def _putblock_raw(self, block, finish):
        """ put the data into the socket """
        pos = 0
        last = 0
        while not last:
            data = memoryview(block)[pos:pos + MAX_PACKAGE_LENGTH]
            length = len(data)
            if length < MAX_PACKAGE_LENGTH:
                last = 1
            flag = struct.pack('<H', (length << 1) + (last if finish else 0))
            self.socket.send(flag)
            self.socket.send(data)
            pos += length

    def _send_all_and_shutdown(self, block):
        """ put the data into the socket """
        pos = 0
        end = len(block)
        block = memoryview(block)
        while pos < end:
            data = block[pos:pos + 8192]
            nsent = self.socket.send(data)
            pos += nsent
        try:
            self.socket.shutdown(socket.SHUT_WR)
        except OSError:
            pass

    def __del__(self):
        if self.socket:
            self.socket.close()

    def set_reply_size(self, size):
        # type: (int) -> None
        """
        Set the amount of rows returned by the server.

        args:
            size: The number of rows
        """

        self.cmd("Xreply_size %s" % size)

    def set_uploader(self, uploader: "Uploader"):
        """Register the given Uploader, or None to deregister"""
        self.uploader = uploader

    def set_downloader(self, downloader: "Downloader"):
        """Register the given Downloader, or None to deregister"""
        self.downloader = downloader


# When all supported Python versions support it we can enable @dataclass here.
class HandshakeOption:
    """
    Option that can be set during the MAPI handshake

    Should be sent as <name>=<val>, where <val> is `value` converted to int.
    The `level` is used to determine if the server supports this option.
    The `fallback` is a function-like object that can be called with the
    value (not converted to an integer) as a parameter.
    Field `sent` can be used to keep track of whether the option has been sent.
    """

    def __init__(self, level, name, fallback, value):
        self.level = level
        self.name = name
        self.value = value
        self.fallback = fallback
        self.sent = False


def construct_target_from_args(database: Optional[str], username: str, password: str, language: str,  # noqa: C901
                               hostname: Optional[str] = None, port: Optional[int] = None, unix_socket: Optional[str] = None,
                               connect_timeout: Optional[Union[float, int]] = None,
                               **kwargs):
    """Construct a Target from the other args"""

    target = Target()

    if database is not None:
        assert isinstance(database, str)
        target.database = database
    if username is not None:
        target.user = username
    if password is not None:
        target.password = password
    if language is not None:
        target.language = language
    if hostname is not None:
        target.host = hostname
    if port is not None:
        target.port = port
    if unix_socket is not None:
        target.sock = unix_socket
    if connect_timeout is not None:
        target.connect_timeout = connect_timeout
    for key, value in kwargs.items():
        try:
            target.set(key, value)
        except ValueError as e:
            raise DatabaseError(f"{key}={value}: {e}")

    return target



#!/usr/bin/env python3

import os
import re
from typing import List, Optional, Tuple
from unittest import TestCase
import unittest

from pymonetdb.target import VIRTUAL, Target, parse_bool


class Line(str):
    """A Line is a string that remembers which file and line number it came from"""
    file: str
    idx: int
    nr: int

    def __new__(cls, text: str, file: str, idx: int):
        line = super().__new__(cls, text)
        line.file = file
        line.idx = idx
        line.nr = idx + 1
        return line

    @property
    def location(self):
        return self.file + ":" + str(self.nr)


def read_lines(f, filename: str, start_line=0) -> List[Line]:
    """Read from 'f' and turn the lines into Lines"""
    n = start_line
    lines = []
    for s in f:
        s = s.rstrip()
        line = Line(s, filename, n)
        lines.append(line)
        n += 1
    return lines


def split_tests(lines: List[Line]) -> List[Tuple[str, List[Line]]]:
    tests: List[Tuple[str, List[Line]]] = []
    cur: Optional[List[Line]] = None
    header = None
    count = 0
    location = None
    for line in lines:
        if cur is None:
            if line.startswith("```test"):
                location = line.location
                cur = []
            elif line.startswith('#'):
                header = line.lstrip('#').strip()
                header = re.sub(r'\W+', '_', header).lower()
                count = 0
        else:
            if line.startswith("```"):
                count += 1
                name = f"{header}_{count}"
                assert len(cur) > 0
                tests.append((name, cur))
                cur = None
            else:
                cur.append(line)
    if cur is not None:
        raise Exception(f"Unclosed block at {location}")
    return tests


# Note: we programmatically add test methods to this file
# based on the contents of tests.md.
class TargetTests(TestCase):

    def test_dummy_tests(self):
        """Convenience method. Run the tests from t.md if that exists."""
        filename = 't.md'
        if not os.path.exists(filename):
            raise unittest.SkipTest(f"{filename} does not exist")
        lines = read_lines(open(filename), filename)
        tests = split_tests(lines)
        for name, test in tests:
            self.run_test(test)

    def run_test(self, test):
        target = Target()
        for line in test:
            try:
                self.apply_line(target, line)
                continue
            except AssertionError as e:
                if hasattr(e, 'add_note'):
                    e.add_note(f"At {line.location}")
                    raise
                else:
                    raise AssertionError(f"At {line.location}: {e}")
            except unittest.SkipTest:
                break
            except Exception as e:
                if hasattr(e, 'add_note'):
                    e.add_note(f"At {line.location}")
                    raise

    def apply_line(self, target: Target, line: Line):  # noqa C901
        if not line:
            return

        command, rest = line.split(None, 1)
        command = command.upper()
        if command == "PARSE":
            self.apply_parse(target, rest)
        elif command == "ACCEPT":
            self.apply_accept(target, rest)
        elif command == "REJECT":
            self.apply_reject(target, rest)
        elif command == "EXPECT":
            key, value = rest.split('=', 1)
            self.apply_expect(target, key, value)
        elif command == "SET":
            key, value = rest.split('=', 1)
            self.apply_set(target, key, value)
        elif command == "ONLY":
            impl = rest
            if impl != 'pymonetdb':
                raise unittest.SkipTest(f"only for {impl}")
        elif command == "NOT":
            impl = rest
            if impl == 'pymonetdb':
                raise unittest.SkipTest(f"not for {impl}")
        else:
            self.fail(f"Unknown command: {command}")

    def apply_parse(self, target: Target, url):
        target.parse(url)

    def apply_accept(self, target: Target, url):
        target.parse(url)
        target.validate()

    def apply_reject(self, target: Target, url):
        try:
            target.parse(url)
        except ValueError:
            return
        # last hope
        try:
            target.validate()
        except ValueError:
            return
        raise ValueError("Expected URL to be rejected")

    def apply_set(self, target: Target, key, value):
        target.set(key, value)

    def apply_expect(self, target: Target, key, expected_value):
        if key == 'valid':
            return self.apply_expect_valid(target, key, expected_value)

        if key in VIRTUAL:
            target.validate()

        if key == 'connect_binary':
            actual_value = target.connect_binary(65535)
        else:
            actual_value = target.get(key)

        self.verify_expected_value(key, expected_value, actual_value)

    def verify_expected_value(self, key, expected_value, actual_value):
        if isinstance(actual_value, bool):
            expected_value = parse_bool(expected_value)
        elif isinstance(actual_value, int):
            try:
                expected_value = int(expected_value)
            except ValueError:
                # will show up in the comparison below
                pass
        if actual_value != expected_value:
            self.fail(f"Expected {key}={expected_value!r}, found {actual_value!r}")

    def apply_expect_valid(self, target, key, expected_value):
        should_succeed = parse_bool(expected_value)
        try:
            target.validate()
            if not should_succeed:
                self.fail("Expected valid=false")
        except ValueError as e:
            if should_succeed:
                self.fail(f"Expected valid=true, got error {e}")
        return


# Magic alert!
# Read tests.md and generate test cases programmatically!
filename = os.path.join(os.path.dirname(__file__), 'tests.md')
lines = read_lines(open(filename), filename)
tests = split_tests(lines)
for name, test in tests:
    if test:
        line_nr = f"line_{test[0].nr}_"
    else:
        line_nr = ""

    def newlexicalscope(test):
        return lambda this: this.run_test(test)
    setattr(TargetTests, f"tests_md_{line_nr}{name}", newlexicalscope(test))



"""
This is the python implementation of the mapi protocol.
"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.


from unittest import TestCase

import pymonetdb
from pymonetdb.exceptions import OperationalError
from tests.util import test_args


class TestContextManager(TestCase):

    def connect(self):
        return pymonetdb.connect(**test_args)

    def test_context_manager(self):
        """Test using a Connection in a with-clause"""

        # This is just a quick example of how connections and cursors
        # would typically be used as a context manager.
        # More detailed test cases are given below.

        with self.connect() as conn, conn.cursor() as c:
            c.execute("SELECT 42")
            self.assertEqual(c.fetchone()[0], 42)

    def test_connection_closes_when_ok(self):
        x = self.connect()
        self.assertIsNotNone(x.mapi)
        try:
            with x as conn:
                y = conn.cursor()
                self.assertIsNotNone(y.connection)
                try:
                    with y as c:
                        c.execute("SELECT 42")
                        self.assertEqual(c.fetchone()[0], 42)
                finally:
                    # check cursor is closed
                    self.assertIsNone(y.connection)
        finally:
            # check connection is closed
            self.assertIsNone(x.mapi)

    def test_connection_closes_when_sql_error(self):
        x = self.connect()
        y = x.cursor()
        self.assertIsNotNone(x.mapi)
        self.assertIsNotNone(y.connection)
        with self.assertRaisesRegex(OperationalError, expected_regex="Unexpected symbol"):
            with x as conn:
                # suppress warning about unused 'conn'
                if conn:
                    pass
                with y as cursor:
                    cursor.execute("SELECT 42zzz")   # This fails
                    self.fail("the statement above should have raised an exception")
        # check cursor and conn have been closed
        self.assertIsNone(y.connection)
        self.assertIsNone(x.mapi)

    def test_connection_closes_when_other_error(self):
        x = self.connect()
        y = x.cursor()
        self.assertIsNotNone(x.mapi)
        self.assertIsNotNone(y.connection)
        with self.assertRaises(ZeroDivisionError):
            with x as conn:
                # suppress warning about unused 'conn'
                if conn:
                    pass
                with y as cursor:
                    # suppress warning about unused 'conn' and 'cursor'
                    if conn or cursor:
                        pass
                    # create a ZeroDivisionError without static analyzers noticing it
                    one = 1
                    zero = 0
                    one / zero
                    self.fail("the statement above should have raised an exception")
        # check cursor and conn have been closed
        self.assertIsNone(y.connection)
        self.assertIsNone(x.mapi)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.


import copy


class BatchPolicy:
    """This class centralizes the decisions about result set batch sizes"""

    SMALL_NUMBER = 10
    DEFAULT_NUMBER = 100
    BIG_NUMBER = 2_500

    MAX_BINARY_LEVEL = 1

    # To be set by the user
    binary_level = MAX_BINARY_LEVEL
    replysize = DEFAULT_NUMBER
    maxprefetch = BIG_NUMBER

    # Determined during handshake
    server_binexport_level = 0

    # per-cursor state
    last = 0

    def __init__(self):
        pass

    def clone(self) -> "BatchPolicy":
        return copy.copy(self)

    def use_binary(self) -> bool:
        return self.binary_level > 0 and self.server_binexport_level > 0

    def _effective_reply_size(self) -> int:
        if self.use_binary() and self.replysize < 0:
            # Only include a few rows in the initial reply so the rest can
            # be fetched using the binary protocol
            return self.SMALL_NUMBER
        else:
            # return replysize even when it's negative
            return self.replysize

    def handshake_reply_size(self) -> int:
        return self._effective_reply_size()

    def decide_arraysize(self) -> int:
        if self.replysize > 0:
            return self.replysize
        else:
            return self.DEFAULT_NUMBER

    def new_query(self) -> int:
        # reply size computation is the same
        reply_size = self._effective_reply_size()
        # Note how we set self.last to reply_size rather than to 0 as we do
        # in .scroll().
        # This is because the initial reply will already have happened
        # by the time we look at it.
        self.last = reply_size
        return reply_size

    def scroll(self):
        # Note how we set self.last to 0 rather than to reply_size as we do
        # in .new_query().
        # This is because scrolling does not cause a fetch, it just disposes
        # the currently cached rows.
        self.last = 0

    def batch_size(self,
                   already_used: int,
                   request_start: int, request_end: int,
                   result_end: int
                   ) -> int:

        assert request_start <= request_end <= result_end

        if self.use_binary() and self.replysize < 0:
            # special case. application wants to retrieve
            # everything in one go but we kept the initial
            # reply small because the binary protocol
            # is more efficient.
            # now retrieve the rest in one go.
            return result_end - request_start

        if self.last > 0:
            size = 2 * self.last
        else:
            size = self._effective_reply_size()
        prefetch_end = request_start + size

        # align to fetchmany stride,
        # exploit that the % operator always returns nonnegative.
        real_start = request_start - already_used
        adjustment = (real_start - prefetch_end) % (request_end - real_start)
        prefetch_end += adjustment

        # apply maxprefetch
        if self.maxprefetch >= 0:
            limit = request_end + self.maxprefetch
            if prefetch_end > limit:
                prefetch_end = limit

        # do not fetch beyond the end
        if prefetch_end > result_end:
            prefetch_end = result_end

        # We have computed how much we would prefetch, but maybe the user
        # explicitly asked for more than that
        end = max(prefetch_end, request_end)
        to_fetch = end - request_start
        self.last = to_fetch
        return to_fetch



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.


from typing import List, Optional
from unittest import TestCase
from urllib.parse import parse_qsl, urlencode, urlparse
import pymonetdb
from pymonetdb.policy import BatchPolicy
from tests.util import test_args, test_url


def run_scenario(rowcount: int,
                 server_binexport_level: int,
                 pattern: List[int],
                 binary: Optional[bool] = None,
                 replysize: Optional[int] = None,
                 maxprefetch: Optional[int] = None,
                 ):

    # simulate connect
    policy = BatchPolicy()
    if binary is not None:
        policy.binary_level = binary
    if replysize is not None:
        policy.replysize = replysize
    if maxprefetch is not None:
        policy.maxprefetch = maxprefetch
    policy.server_binexport_level = server_binexport_level

    handshake_reply_size = policy.handshake_reply_size()

    # simulate new cursor
    pol = policy.clone()
    array_size = pol.decide_arraysize()

    # simulate execute
    query_reply_size = pol.new_query()
    pos = 0
    cache_start = 0
    if query_reply_size > 0:
        cache_end = query_reply_size
    else:
        cache_end = rowcount
    intervals = [(cache_start, cache_end)]

    # simulate fetch*()
    for size in pattern:
        if size < 0:
            # simulate fetchall
            size = rowcount - pos
        requested_end = min(pos + size, rowcount)

        # take existing rows
        end = min(requested_end, cache_end)
        existing_rows = end - pos
        pos = end

        if pos == requested_end:
            continue

        fetch_size = pol.batch_size(existing_rows, pos, requested_end, rowcount)
        end = pos + fetch_size
        intervals.append((pos, end))
        cache_start = pos
        cache_end = end
        pos = requested_end

    return Scenario(handshake_reply_size, array_size, query_reply_size, intervals)


class Scenario:
    handshake_reply_size: int
    array_size: int
    query_reply_size: int
    intervals: List[int]

    def __init__(self, handshake_reply_size, array_size, query_reply_size, intervals):
        self.handshake_reply_size = handshake_reply_size
        self.array_size = array_size
        self.query_reply_size = query_reply_size
        self.intervals = intervals

    def __eq__(self, other):
        return (self.handshake_reply_size == other.handshake_reply_size
                and self.array_size == other.array_size
                and self.query_reply_size == other.query_reply_size
                and self.intervals == other.intervals)

    def __repr__(self):
        return (
            "Scenario("
            f"handshake_reply_size={self.handshake_reply_size}, "
            f"array_size={self.array_size}, "
            f"query_reply_size={self.query_reply_size}, "
            f"intervals={self.intervals})"
        )


class TestBatchPolicy(TestCase):
    """Test the BatchPolicy in isolation"""

    def setUp(self):
        self.addTypeEqualityFunc(Scenario, self.compare_scenarios)

    def compare_scenarios(self, left, right, msg=None):
        if left.handshake_reply_size != right.handshake_reply_size:
            self.failureException(
                "scenario differs in handshake_reply_size: "
                f"{left.handshake_reply_size} vs. {right.handshake_reply_size}")
        if left.array_size != right.array_size:
            self.failureException(
                "scenario differs in array_size: "
                f"{left.array_size} vs. {right.array_size}")
        if left.query_reply_size != right.query_reply_size:
            self.failureException(
                "scenario differs in query_reply_size: "
                f"{left.query_reply_size} vs. {right.query_reply_size}")
        if left.intervals != right.intervals:
            try:
                self.assertListEqual(left.intervals, right.intervals)
            except AssertionError as e:
                raise self.failureException(f"scenario differs in intervals: {e}")
            raise self.failureException(f"scenario differs in intervals: {left.intervals} vs. {right.intervals}")

    def run_both(self,
                 rowcount: int,
                 pattern: List[int],
                 binary: Optional[bool] = None,
                 replysize: Optional[int] = None,
                 maxprefetch: Optional[int] = None,
                 ):
        without_binary = run_scenario(
            rowcount, 0, pattern,
            binary=binary, replysize=replysize, maxprefetch=maxprefetch)
        with_binary = run_scenario(
            rowcount, 1, pattern,
            binary=binary, replysize=replysize, maxprefetch=maxprefetch)
        with_binary_but_disabled = run_scenario(
            rowcount, 1, pattern,
            binary=False, replysize=replysize, maxprefetch=maxprefetch)

        try:
            self.assertEqual(without_binary, with_binary)
        except AssertionError as e:
            raise AssertionError("Scenarios differ between binary=False and binary=True") from e

        try:
            self.assertEqual(without_binary, with_binary_but_disabled)
        except AssertionError as e:
            raise AssertionError("Behavior of binary=False and server_support_binary=False differs") from e

        return without_binary

    def test_defaults_fetchall(self):
        # the first 100 rows have already been fetched, fetchall fetches the rest.
        scen = self.run_both(1000, [-1])
        self.assertEqual(100, scen.handshake_reply_size)
        self.assertEqual(100, scen.array_size)
        self.assertEqual(100, scen.query_reply_size)
        self.assertEqual(
            [(0, 100), (100, 1000)],
            scen.intervals)

    def test_defaults_fetchone(self):
        # if we request the rows one by one we get the typical doubling batch size.
        scen = self.run_both(1000, 1000 * [1])
        self.assertEqual(100, scen.handshake_reply_size)
        self.assertEqual(100, scen.array_size)
        self.assertEqual(100, scen.query_reply_size)
        self.assertEqual(
            [(0, 100), (100, 300), (300, 700), (700, 1000)],
            scen.intervals)

    def test_fetchmany_aligned(self):
        # if the stride fits the reply size there is no difference with fetchone
        scen = self.run_both(1000, 200 * [50])
        self.assertEqual(100, scen.handshake_reply_size)
        self.assertEqual(100, scen.array_size)
        self.assertEqual(100, scen.query_reply_size)
        self.assertEqual(
            [(0, 100), (100, 300), (300, 700), (700, 1000)],
            scen.intervals)

    def test_stride(self):
        # with fetchmany, the batch sizes adjust to the stride.
        scen = self.run_both(1000, 8 * [40])
        self.assertEqual(100, scen.handshake_reply_size)
        self.assertEqual(100, scen.array_size)
        self.assertEqual(100, scen.query_reply_size)
        self.assertEqual([
            (0, 100),       # initial fetch
            (100, 320),     # 320 instead of 300
        ], scen.intervals)

    def test_fetchmany_unaligned(self):
        scen = self.run_both(1000, 100 * [75])
        self.assertEqual(100, scen.handshake_reply_size)
        self.assertEqual(100, scen.array_size)
        self.assertEqual(100, scen.query_reply_size)
        self.assertEqual([
            (0, 100),  # initial fetch
            (100, 300),  # 300 is a multiple of 75
            (300, 750),  # 700 is not, but 750 is
            (750, 1000)
        ], scen.intervals)

    def test_small_reply_size(self):
        # fetchall behaves as expected
        scen = self.run_both(1000, [-1], replysize=20)
        self.assertEqual(20, scen.handshake_reply_size)
        self.assertEqual(20, scen.array_size)
        self.assertEqual(20, scen.query_reply_size)
        self.assertEqual(
            [(0, 20), (20, 1000)],
            scen.intervals)

        # batches still double but they start smaller
        scen = self.run_both(1000, 1000 * [1], replysize=20)
        self.assertEqual(20, scen.handshake_reply_size)
        self.assertEqual(20, scen.array_size)
        self.assertEqual(20, scen.query_reply_size)
        self.assertEqual([
            (0, 20),
            (20, 60),
            (60, 140),
            (140, 300),
            (300, 620),
            (620, 1000)
        ], scen.intervals)

    def test_small_prefetch(self):
        # batch does not grow larger than stride + prefetch = 100 + 100 = 200
        scen = self.run_both(1000, 100 * [100], maxprefetch=100)
        self.assertEqual(100, scen.handshake_reply_size)
        self.assertEqual(100, scen.array_size)
        self.assertEqual(100, scen.query_reply_size)
        self.assertEqual([
            (0, 100),
            (100, 300),
            (300, 500),
            (500, 700),
            (700, 900),
            (900, 1000)
        ], scen.intervals)

    def test_small_prefetch_large_stride(self):
        # if maxprefetch is small and does not match the stride it ends
        # up not really being used
        scen = self.run_both(1000, 30 * [250], maxprefetch=100)
        self.assertEqual(100, scen.handshake_reply_size)
        self.assertEqual(100, scen.array_size)
        self.assertEqual(100, scen.query_reply_size)
        self.assertEqual([
            (0, 100),
            (100, 350),  # steps of 250, but shifted
            (350, 600),
            (600, 850),
            (850, 1000),
        ], scen.intervals)

    def test_no_prefetch_large_stride(self):
        # without prefetch it goes 100, 150, 250, 250, ...
        scen = self.run_both(1000, 30 * [250], maxprefetch=0)
        self.assertEqual(100, scen.handshake_reply_size)
        self.assertEqual(100, scen.array_size)
        self.assertEqual(100, scen.query_reply_size)
        self.assertEqual([
            (0, 100),
            (100, 250),  # complete the first step
            (250, 500),  # fetch the remaining steps individually
            (500, 750),
            (750, 1000),
        ], scen.intervals)

    def test_unlimited_prefetch(self):
        # if we turn off the prefetch limit it grows without bound
        scen = self.run_both(1_000_000, 10_000 * [100], maxprefetch=-1)
        self.assertEqual(100, scen.handshake_reply_size)
        self.assertEqual(100, scen.array_size)
        self.assertEqual(100, scen.query_reply_size)
        self.assertEqual([
            (0, 100),
            (100, 300),
            (300, 700),
            (700, 1500),
            (1500, 3100),
            (3100, 6300),
            (6300, 12700),
            (12700, 25500),
            (25500, 51100),
            (51100, 102300),
            (102300, 204700),
            (204700, 409500),
            (409500, 819100),
            (819100, 1000000)
        ], scen.intervals)

    def test_unlimited_replysize_no_binary(self):
        # everything comes in the initial response
        scen = run_scenario(1000, 0, 10 * [100], replysize=-1)
        self.assertEqual(-1, scen.handshake_reply_size)
        self.assertEqual(100, scen.array_size)
        self.assertEqual(-1, scen.query_reply_size)
        self.assertEqual([(0, 1000)], scen.intervals)

        # same if binary is possible but disabled
        scen = run_scenario(1000, 1, 10 * [100], replysize=-1, binary=False)
        self.assertEqual(-1, scen.handshake_reply_size)
        self.assertEqual(100, scen.array_size)
        self.assertEqual(-1, scen.query_reply_size)
        self.assertEqual([(0, 1000)], scen.intervals)

    def test_unlimited_replysize_with_binary(self):
        # we set replysize to -1 but it still sends 10 to the server
        # to keep the initial response small and retrieve
        # rest using the binary protocol
        scen = run_scenario(1000, 1, 10 * [100], replysize=-1)
        self.assertEqual(10, scen.handshake_reply_size)
        self.assertEqual(100, scen.array_size)
        self.assertEqual(10, scen.query_reply_size)
        self.assertEqual([(0, 10), (10, 1000)], scen.intervals)

    def test_arraysize(self):
        # arraysize follows the replysize of the connection
        # at the time the cursor was created
        master = BatchPolicy()
        master.replysize = 50
        pol = master.clone()
        self.assertEqual(50, pol.decide_arraysize())

        master.replysize = -1
        pol = master.clone()
        self.assertEqual(100, pol.decide_arraysize())


class TestPolicySetting(TestCase):
    _server_has_binary: Optional[bool]
    _conns: List[pymonetdb.Connection]

    def setUp(self):
        self._conns = []

    def tearDown(self):
        for conn in self._conns:
            try:
                conn.close()
            except pymonetdb.exceptions.Error:
                pass
        self._conns = []

    def _connect(self, **kw_args):
        try:
            args = dict()
            args.update(test_args)
            for key in ['replysize', 'maxprefetch', 'binary']:
                if key in args:
                    del args[key]
            args.update(kw_args)
            conn = pymonetdb.connect(**args)
            self._conns.append(conn)
            return conn
        except AttributeError:
            self.fail("No connect method found in pymonetdb module")

    def check_more(self, conn: pymonetdb.Connection):
        self.assertEqual(conn.binary > 0, conn._policy.binary_level)
        self.assertEqual(conn.replysize, conn._policy.replysize)
        self.assertEqual(conn.maxprefetch, conn._policy.maxprefetch)

        cursor = conn.cursor()

        self.assertEqual(cursor.binary > 0, conn.binary)
        self.assertEqual(cursor.replysize, conn.replysize)
        self.assertEqual(cursor.maxprefetch, conn.maxprefetch)
        if cursor.replysize < 0:
            self.assertEqual(100, cursor.arraysize)
        else:
            self.assertEqual(cursor.replysize, cursor.arraysize)

        self.assertEqual(cursor.binary > 0, cursor._policy.binary_level)
        self.assertEqual(cursor.replysize, cursor._policy.replysize)
        self.assertEqual(cursor.maxprefetch, cursor._policy.maxprefetch)

    def test_defaults(self):
        conn = self._connect()
        self.assertEqual(1, conn._policy.binary_level)   # by default we WANT binary
        self.assertEqual(100, conn._policy.replysize)
        self.assertEqual(2_500, conn._policy.maxprefetch)

        self.check_more(conn)

    def test_constructor_parameters1(self):
        conn = self._connect(replysize=99, maxprefetch=333, binary=0)
        self.assertEqual(0, conn._policy.binary_level)
        self.assertEqual(99, conn._policy.replysize)
        self.assertEqual(333, conn._policy.maxprefetch)

        self.check_more(conn)

    def test_constructor_parameters2(self):
        conn = self._connect(replysize=-1, maxprefetch=-1, binary=1)
        self.assertEqual(1, conn._policy.binary_level)
        self.assertEqual(-1, conn._policy.replysize)
        self.assertEqual(-1, conn._policy.maxprefetch)

        self.check_more(conn)

    def test_conn_attr_updates(self):
        conn = self._connect()
        conn.replysize = 99
        conn.maxprefetch = 333
        conn.binary = 0
        self.assertEqual(0, conn._policy.binary_level)
        self.assertEqual(99, conn._policy.replysize)
        self.assertEqual(333, conn._policy.maxprefetch)

        self.check_more(conn)

    def test_cursor_attr_updates(self):
        conn = self._connect()
        cursor = conn.cursor()
        cursor.replysize = 99
        cursor.maxprefetch = 333
        cursor.binary = 0
        self.assertEqual(0, cursor._policy.binary_level)
        self.assertEqual(99, cursor._policy.replysize)
        self.assertEqual(333, cursor._policy.maxprefetch)

    def update_url(self, replysize, maxprefetch, binary) -> str:
        u = urlparse(test_url)
        opts = dict(parse_qsl(u.query))
        opts['replysize'] = replysize
        opts['maxprefetch'] = maxprefetch
        opts['binary'] = binary
        u = u._replace(query=urlencode(opts))
        return u.geturl()

    def test_url_parameters1(self):
        updated_url = self.update_url(replysize='99', maxprefetch='444', binary='1')
        with pymonetdb.connect(updated_url) as conn:
            self.assertEqual(1, conn._policy.binary_level)
            self.assertEqual(99, conn._policy.replysize)
            self.assertEqual(444, conn._policy.maxprefetch)

    def test_url_parameters2(self):
        updated_url = self.update_url(replysize='999', maxprefetch='44', binary='0')
        with pymonetdb.connect(updated_url) as conn:
            self.assertEqual(0, conn._policy.binary_level)
            self.assertEqual(999, conn._policy.replysize)
            self.assertEqual(44, conn._policy.maxprefetch)



"""
This is a MonetDB Python API.

To set up a connection use pymonetdb.connect()

"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

from typing import Optional
from pymonetdb import sql
from pymonetdb import mapi
from pymonetdb import exceptions

from pymonetdb.profiler import ProfilerConnection
from pymonetdb.sql.connections import Connection
from pymonetdb.sql.pythonize import BINARY, Binary, DATE, Date, Time, Timestamp, DateFromTicks, TimestampFromTicks, \
    TimeFromTicks, NUMBER, ROWID, STRING, TIME, types, DATETIME, TimeTzFromTicks, TimestampTzFromTicks
from pymonetdb.exceptions import Error, DataError, DatabaseError, IntegrityError, InterfaceError, InternalError, \
    NotSupportedError, OperationalError, ProgrammingError, Warning
from pymonetdb.filetransfer.downloads import Download, Downloader
from pymonetdb.filetransfer.uploads import Upload, Uploader
from pymonetdb.filetransfer.directoryhandler import SafeDirectoryHandler
from pymonetdb.target import Target, looks_like_url

__version__ = '1.8.0'

apilevel = "2.0"
threadsafety = 1

paramstyle = "pyformat"   # with sufficiently recent MonetDB versions you can override this to 'named'

__all__ = ['sql', 'mapi', 'exceptions', 'BINARY', 'Binary', 'connect', 'Connection', 'DATE', 'Date', 'Time',
           'Timestamp', 'DateFromTicks', 'TimeFromTicks', 'TimestampFromTicks', 'DataError', 'DatabaseError', 'Error',
           'IntegrityError', 'InterfaceError', 'InternalError', 'NUMBER', 'NotSupportedError', 'OperationalError',
           'ProgrammingError', 'ROWID', 'STRING', 'TIME', 'Warning', 'apilevel', 'connect', 'paramstyle',
           'threadsafety', 'Download', 'Downloader', 'Upload', 'Uploader', 'SafeDirectoryHandler', 'types', 'DATETIME',
           'TimeTzFromTicks', 'TimestampTzFromTicks']


def connect(    # noqa C901
        database: str,
        hostname: Optional[str] = None,
        port: Optional[int] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        unix_socket: Optional[str] = None,
        autocommit: Optional[bool] = None,
        host: Optional[str] = None,
        user: Optional[str] = None,
        connect_timeout: Optional[int] = None,
        binary: Optional[int] = None,
        replysize: Optional[str] = None,
        maxprefetch: Optional[str] = None,
        *,
        sock: Optional[str] = None,
        sockdir: Optional[str] = None,
        tls: Optional[bool] = None,
        cert: Optional[str] = None,
        certhash: Optional[str] = None,
        clientkey: Optional[str] = None,
        clientcert: Optional[str] = None,
        schema: Optional[str] = None,
        timezone: Optional[int] = None,
        dangerous_tls_nocheck: Optional[str] = None,
):
    """Set up a connection to a MonetDB SQL database

    Can connect using TCP or using a Unix Domain socket.

    database : str
        name of the database, or a
        `MonetDB URL <https://www.monetdb.org/documentation/user-guide/client-interfaces/monetdb-urls/>`_
    hostname : str
        host name to make TCP connection to
    port : int
        port number to connect to (default: 50000), also used to
        construct default Unix Domain socket name
    username : str
        user name to authenticate as (default: monetdb)
    password : str
        password to authenticate with (default: monetdb)
    unix_socket : str
        Unix Domain socket to connect to when host name is not set
        (default: /tmp/.s.monetdb.PORTNUMBER)
    autocommit : bool
        enable/disable auto commit (default: false)
    host : str
        alias for hostname
    user : str
        alias for username
    connect_timeout : float
        socket timeout when connecting, in seconds
    binary : int
        enable binary result sets when possible if > 0 (default: 1)
    replysize : str
        number of rows to retrieve immediately after query execution (default: 100, -1 means everything)
    maxprefetch : str
        max. number of additional rows to fetch during Cursor.fetchone() or Cursor.fetchmany()
    sock : str
        alias for unix_socket
    sockdir: str
        directory where Unix Domain sockets are searched (default: /tmp)
    tls : bool
        whether to secure (encrypt) the connection
    cert : str
        optional path to TLS certificate to verify the server with
    certhash : str
        if given, only verify that server certificate has this fingerprint, implies dangerous_tls_nocheck=host,cert.
        format: sha256:digits
    clientkey : str
        optional path to TLS key to present to server for authentication
    clientcert : str
        optional path to TLS cert to present to server for authentication,
        this can also be appended to the key file.
    schema : str
        the schema to select after connecting
    timezone : int
        the time zone to use, in minutes east of UTC
    dangerous_tls_nocheck : str
        comma-separated list of certificate checks to skip during connecting:
        'host': ignore host name mismatch,
        'cert': ignore bad certificate chain
    """

    target = Target()

    if hostname is not None:
        target.host = hostname
    if port is not None:
        target.port = port
    if username is not None:
        target.user = username
    if password is not None:
        target.password = password
    if unix_socket is not None:
        target.sock = unix_socket
    if autocommit is not None:
        target.autocommit = autocommit
    if host is not None:
        target.host = host
    if user is not None:
        target.user = user
    if connect_timeout is not None:
        target.connect_timeout = connect_timeout
    if binary is not None:
        target.binary = binary
    if replysize is not None:
        target.replysize = replysize
    if maxprefetch is not None:
        target.maxprefetch = maxprefetch
    if sock is not None:
        target.sock = sock
    if sockdir is not None:
        target.sockdir = sockdir
    if tls is not None:
        target.tls = tls
    if cert is not None:
        target.cert = cert
    if certhash is not None:
        target.certhash = certhash
    if clientkey is not None:
        target.clientkey = clientkey
    if clientcert is not None:
        target.clientcert = clientcert
    if schema is not None:
        target.schema = schema
    if timezone is not None:
        target.timezone = timezone
    if dangerous_tls_nocheck is not None:
        target.dangerous_tls_nocheck = dangerous_tls_nocheck

    if looks_like_url(database):
        target.boundary()
        target.parse(database)
    else:
        target.database = database

    return Connection(target)


def profiler_connection(*args, **kwargs):
    c = ProfilerConnection()
    c.connect(*args, **kwargs)
    return c


profiler_connection.__doc__ = ProfilerConnection.__init__.__doc__



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.


from unittest import SkipTest, TestCase

from pymonetdb import OperationalError
import pymonetdb
from tests.util import test_args


class TestNamed(TestCase):

    def connect(self):
        return pymonetdb.connect(**test_args)

    def setUp(self):
        with self.connect() as conn, conn.cursor() as c:
            # does it work at all?
            c.execute("SELECT 42")
            # does it support named arguments?
            try:
                c.execute("SELECT :fortytwo : ( fortytwo 42 )")
            except OperationalError:
                raise SkipTest("Named parameter syntax not supported by this MonetDB")

        # if we get here, temporarily change paramstyle
        assert pymonetdb.paramstyle == 'pyformat'
        pymonetdb.paramstyle = 'named'

    def tearDown(self):
        pymonetdb.paramstyle = 'pyformat'

    def test_named_parameters(self):
        with self.connect() as conn, conn.cursor() as c:
            parms = dict(foo=42, bar="banana")
            c.execute("SELECT :foo AS foo, :bar AS bar", parms)
            row = c.fetchone()
            self.assertEqual(row, (42, "banana"))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

"""
functions for converting binary result sets to Python objects
"""

from abc import abstractmethod
import array
from datetime import date, datetime, time, timezone, timedelta
from decimal import Decimal
import json
from math import isnan
import struct
import sys
from typing import Any, Callable, List, Optional
from uuid import UUID
from pymonetdb.exceptions import InternalError

from pymonetdb.sql import types
import pymonetdb.sql.cursors


INT_WIDTH_TO_ARRAY_TYPE = {}
for code in 'bhilq':
    bit_width = 8 * array.array(code).itemsize
    INT_WIDTH_TO_ARRAY_TYPE[bit_width] = code

FLOAT_WIDTH_TO_ARRAY_TYPE = {}
for code in 'fd':
    bit_width = 8 * array.array(code).itemsize
    FLOAT_WIDTH_TO_ARRAY_TYPE[bit_width] = code

# very unlikely but if we ever encounter a Python with highly unstandard
# float sizes we want to know
assert FLOAT_WIDTH_TO_ARRAY_TYPE[32] == 'f'
assert FLOAT_WIDTH_TO_ARRAY_TYPE[64] == 'd'


class BinaryDecoder:
    @abstractmethod
    def decode(self, server_endian: str, data: memoryview) -> List[Any]:
        """Interpret the given bytes as a list of Python objects"""
        pass


class IntegerDecoder(BinaryDecoder):
    array_letter: str
    null_value: int
    mapper: Optional[Callable[[int], Any]]

    def __init__(self,
                 width: int,
                 mapper: Optional[Callable[[int], Any]] = None):
        self.mapper = mapper
        self.array_letter = INT_WIDTH_TO_ARRAY_TYPE[width]
        self.null_value = -(1 << (width - 1))

    def decode(self, server_endian: str, data: memoryview) -> List[Any]:
        arr = array.array(self.array_letter)
        arr.frombytes(data)
        if server_endian != sys.byteorder:
            arr.byteswap()
        null_value = self.null_value
        if self.mapper:
            m = self.mapper
            values = [None if v == null_value else m(v) for v in arr]
        else:
            values = [None if v == null_value else v for v in arr]
        return values


class HugeIntDecoder(BinaryDecoder):
    mapper: Optional[Callable[[int], Any]]

    def __init__(self, mapper: Optional[Callable[[int], Any]] = None):
        self.mapper = mapper

    def decode(self, server_endian: str, data: memoryview) -> List[Any]:
        # we cannot directly decode 128 bits but we can decode 64 bits
        letter = INT_WIDTH_TO_ARRAY_TYPE[64].upper()
        arr = array.array(letter)
        arr.frombytes(data)
        if server_endian != sys.byteorder:
            arr.byteswap()
        # maybe some day we can come up with something faster
        result: List[Optional[int]] = []
        high1 = 1 << 64
        null_value = 1 << 127
        wrap = 1 << 128
        (hi_idx, lo_idx) = (0, 1) if server_endian == 'big' else (1, 0)
        if self.mapper is None:
            for i in range(0, len(arr), 2):
                hi = arr[i + hi_idx]
                lo = arr[i + lo_idx]
                n = high1 * hi + lo
                if n == null_value:
                    result.append(None)
                elif n >= null_value:
                    result.append(n - wrap)
                else:
                    result.append(n)
        else:
            mapper = self.mapper
            for i in range(0, len(arr), 2):
                hi = arr[i + hi_idx]
                lo = arr[i + lo_idx]
                n = high1 * hi + lo
                if n == null_value:
                    result.append(None)
                elif n >= null_value:
                    result.append(mapper(n - wrap))
                else:
                    result.append(mapper(n))
        return result


class FloatDecoder(BinaryDecoder):
    array_letter: str

    def __init__(self, width: int):
        self.array_letter = FLOAT_WIDTH_TO_ARRAY_TYPE[width]

    def decode(self, server_endian: str, data: memoryview) -> List[Any]:
        arr = array.array(self.array_letter)
        arr.frombytes(data)
        if server_endian != sys.byteorder:
            arr.byteswap()
        values = [None if isnan(v) else v for v in arr]
        return values


class UuidDecoder(BinaryDecoder):
    def decode(self, server_endian: str, data: memoryview) -> List[Any]:
        result = []
        null_value = UUID(bytes=16 * b'\x00')
        for i in range(0, len(data), 16):
            u: Optional[UUID] = UUID(bytes=bytes(data[i:i + 16]))
            if u == null_value:
                u = None
            result.append(u)
        return result


def _decode_utf8(x: bytes) -> str:
    return str(x, 'utf-8')


class ZeroDelimitedDecoder(BinaryDecoder):
    def __init__(self, converter: Callable[[bytes], Any]):
        self.converter = converter

    def decode(self, _wrong_endian, data: memoryview) -> List[Any]:
        null_value = b'\x80'
        # tobytes causes a copy but I don't see how that can be avoided
        parts = data.tobytes().split(b'\x00')
        parts.pop()  # empty tail element caused by trailing \x00
        conv = self.converter
        values = [None if v == null_value else conv(v) for v in parts]
        return values


class TimestampDecoder(BinaryDecoder):
    seconds_east: Optional[int]

    def __init__(self, seconds_east: Optional[int]):
        self.seconds_east = seconds_east

    def decode(self, server_endian: str, data: memoryview) -> List[Any]:
        if self.seconds_east is None:
            utczone = None
            ourzone = None
        else:
            utczone = timezone.utc
            ourzone = timezone(timedelta(seconds=self.seconds_east))
        # this is going to be dreadfully slow
        result = []

        for i in range(0, len(data), 12):
            rec = data[i:i + 12]
            micros = int.from_bytes(rec[0:4], server_endian, signed=False)   # type: ignore
            second = int(rec[4])
            minute = int(rec[5])
            hour = int(rec[6])
            # byte 7 is padding
            day = int(rec[8])
            month = int(rec[9])
            year = int.from_bytes(rec[10:12], server_endian, signed=True)   # type: ignore
            if micros < 1e6:
                ts = datetime(year, month, day, hour, minute, second, micros, tzinfo=utczone)
                if ourzone:
                    ts = ts.astimezone(ourzone)
            else:
                ts = None
            result.append(ts)

        return result


class TimeDecoder(BinaryDecoder):
    seconds_east: Optional[int]

    def __init__(self, seconds_east: Optional[int]):
        self.seconds_east = seconds_east

    def decode(self, server_endian: str, data: memoryview) -> List[Any]:
        if self.seconds_east is None:
            ourzone = None
            delta = 0
        else:
            ourzone = timezone(timedelta(seconds=self.seconds_east))
            delta = int(timedelta(seconds=self.seconds_east).total_seconds())
        # this is going to be dreadfully slow
        result = []

        for i in range(0, len(data), 8):
            rec = data[i:i + 12]
            micros = int.from_bytes(rec[0:4], server_endian, signed=False)   # type: ignore
            second = int(rec[4])
            minute = int(rec[5])
            hour = int(rec[6])
            # byte 7 is padding
            if micros < 1e6:
                if delta:
                    adjusted = (3600 * hour + 60 * minute + second + delta) % 86400
                    hour = adjusted // 3600
                    minute = (adjusted // 60) % 60
                    second = adjusted % 60
                t = time(hour, minute, second, micros, tzinfo=ourzone)
            else:
                t = None
            result.append(t)

        return result


class DateDecoder(BinaryDecoder):
    def decode(self, server_endian: str, data: memoryview) -> List[Any]:
        result = []

        for i in range(0, len(data), 4):
            rec = data[i:i + 4]
            day = int(rec[0])
            month = int(rec[1])
            year = int.from_bytes(rec[2:4], server_endian, signed=True)   # type: ignore
            if month <= 12:
                d = date(year, month, day)
            else:
                d = None
            result.append(d)

        return result


class BlobDecoder(BinaryDecoder):
    def decode(self, server_endian: str, data: memoryview) -> List[Any]:
        result: List[Optional[bytes]] = []
        pos = 0

        assert server_endian in ['big', 'little']
        struct_letter = '>q' if server_endian == 'big' else '<q'

        length = len(data)
        while pos < length:
            if 8 > length - pos:
                raise InternalError(f"incomplete blob header after {len(result)} blobs")
            header = struct.unpack(struct_letter, data[pos:pos + 8])[0]
            pos += 8
            if header >= 0:
                if header > length - pos:
                    raise InternalError(f"incomplete blob after {len(result)} blobs")
                result.append(bytes(data[pos:pos + header]))
                pos += header
            else:
                result.append(None)

        return result


def get_decoder(cursor: 'pymonetdb.sql.cursors.Cursor', colno: int) -> Optional[BinaryDecoder]:
    assert cursor.description
    description = cursor.description[colno]
    type_code = description.type_code
    mapper = mapping.get(type_code)
    if not mapper:
        return None
    decoder = mapper(cursor, colno)
    return decoder


def make_decimal_decoder(cursor: 'pymonetdb.sql.cursors.Cursor', colno: int) -> BinaryDecoder:
    assert cursor.description
    description: 'pymonetdb.sql.cursors.Description' = cursor.description[colno]
    scale = 10 ** description.scale
    precision = description.precision

    def mapper(n):
        return Decimal(n) / scale

    if precision <= 2:
        bit_width = 8
    elif precision <= 4:
        bit_width = 16
    elif precision <= 9:
        bit_width = 32
    elif precision <= 18:
        bit_width = 64
    elif precision <= 38:
        # IntegerDecoder doesn't support 128
        return HugeIntDecoder(mapper=mapper)
        bit_width = 128
    else:
        # as far as we know MonetDB only supports up to 38
        assert precision <= 38

    return IntegerDecoder(bit_width, mapper=mapper)


mapping = {
    types.TINYINT: lambda cursor, colno: IntegerDecoder(8),
    types.SMALLINT: lambda cursor, colno: IntegerDecoder(16),
    types.INT: lambda cursor, colno: IntegerDecoder(32),
    types.BIGINT: lambda cursor, colno: IntegerDecoder(64),
    types.HUGEINT: lambda cursor, colno: HugeIntDecoder(),

    types.REAL: lambda cursor, colno: FloatDecoder(32),
    types.FLOAT: lambda cursor, colno: FloatDecoder(64),  # MonetDB defines FLOAT to be 64 bits
    types.DOUBLE: lambda cursor, colno: FloatDecoder(64),

    types.BOOLEAN: lambda cursor, colno: IntegerDecoder(8, mapper=bool),

    types.UUID: lambda cursor, colno: UuidDecoder(),

    types.DECIMAL: make_decimal_decoder,

    types.CHAR: lambda cursor, colno: ZeroDelimitedDecoder(_decode_utf8),
    types.VARCHAR: lambda cursor, colno: ZeroDelimitedDecoder(_decode_utf8),
    types.CLOB: lambda cursor, colno: ZeroDelimitedDecoder(_decode_utf8),
    types.URL: lambda cursor, colno: ZeroDelimitedDecoder(_decode_utf8),
    types.JSON: lambda cursor, colno: ZeroDelimitedDecoder(json.loads),

    types.BLOB: lambda cursor, colno: BlobDecoder(),

    types.TIMESTAMP: lambda cursor, colno: TimestampDecoder(None),
    types.TIMESTAMPTZ: lambda cursor, colno: TimestampDecoder(cursor.connection._current_timezone_seconds_east),
    types.DATE: lambda cursor, colno: DateDecoder(),
    types.TIME: lambda cursor, colno: TimeDecoder(None),
    types.TIMETZ: lambda cursor, colno: TimeDecoder(cursor.connection._current_timezone_seconds_east),

    types.MONTH_INTERVAL: lambda cursor, colno: IntegerDecoder(32),
    types.SEC_INTERVAL: lambda cursor, colno: IntegerDecoder(64, mapper=lambda x: timedelta(milliseconds=x)),
    types.DAY_INTERVAL: lambda cursor, colno: IntegerDecoder(64, mapper=lambda x: timedelta(milliseconds=x).days),

    # Not supported in COPY BINARY or the binary protocol
    # types.GEOMETRY: strip,
    # types.GEOMETRYA: strip,
    # types.INET: str,
    # types.MBR: strip,
    # types.OID: oid,
    # types.XML: str,

    # These are mentioned in pythonize.py but as far as I know the server never
    # produces them
    #
    # types.SERIAL: int,
    # types.SHORTINT: int,
    # types.MEDIUMINT: int,
    # types.LONGINT: int,
    # types.WRD: int,


}



"""
Module containing shared utilities only used for testing MonetDB
"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

import pymonetdb

from importlib import import_module
from os import environ
from urllib.parse import urlencode


def _delete_none_values(d):
    for k in [k for k, v in d.items() if v is None]:
        del d[k]
    return d


def _parse_mapi_env():
    """Read the environment and return a dict that can be passed to mapi.Connection.connect()"""

    # this one is an int and it does not follow the 'TST*' scheme.
    port = int(environ.get('MAPIPORT', '50000'))

    args = {
        'port': port,
        'database': environ.get('TSTDB', 'demo'),
        'hostname': environ.get('TSTHOSTNAME', 'localhost'),
        'username': environ.get('TSTUSERNAME', 'monetdb'),
        'password': environ.get('TSTPASSWORD', 'monetdb'),
        'tls': environ.get('TSTTLS'),
        'cert': environ.get('TSTSERVERCERT'),
    }
    return _delete_none_values(args)


def _parse_control_env():
    """Read the environment and return a dict that can be passed to control.Control()"""

    args = _parse_mapi_env()
    try:
        del args['password']
    except Exception:
        pass
    args['passphrase'] = environ.get('TSTPASSPHRASE', 'testdb')
    return _delete_none_values(args)


def _parse_sql_env():
    """Read the environment and return a dict that can be passed to pymonetdb.connect()"""

    args = _parse_mapi_env()
    args['replysize'] = environ.get('TSTREPLYSIZE')
    args['maxprefetch'] = environ.get('TSTMAXPREFETCH')
    args['binary'] = environ.get('TSTBINARY')
    return _delete_none_values(args)


def _construct_url():
    """Construct a URL version of the result of _parse_sql_env()"""

    # This code is crude. We could use the more sophisticated code of
    # the Target class but we're trying to test that so we'd better not use it
    # for the test itself.
    args = _parse_sql_env()
    for old, new in dict(username='user', hostname='host').items():
        if old in args:
            if new not in args:
                args[new] = args[old]
            del args[old]

    if ':' in args.get('database', ''):
        url = args.get('database', '')
    else:
        bool_strings = dict(
            true=True, false=False,
            yes=True, no=False,
            on=True, off=False)
        tls_arg = args.get('tls', 'off').lower()
        if tls_arg not in bool_strings:
            raise ValueError("TSTTLS must be yes/no/true/false/on/off")
        url = 'monetdbs://' if bool_strings[tls_arg] else 'monetdb://'
        url += args.get('host', args.get('hostname', 'localhost'))
        if args.get('port'):
            url += f":{args['port']}"
        url += '/'
        url += args.get('database', '')
    parms = dict(
        (k, v)
        for k, v in args.items()
        if k not in ['tls', 'host', 'port', 'database']
    )
    if parms:
        sep = '?' if '?' not in url else '&'
        url += sep
        url += urlencode(parms)
    return url


# These are used by the tests to connect to MonetDB:
test_mapi_args = _parse_mapi_env()
control_test_args = _parse_control_env()
test_args = _parse_sql_env()
test_url = _construct_url()


# Control the selection of tests
test_full = environ.get('TSTFULL', 'false').lower() == 'true'
test_control = environ.get('TSTCONTROL', 'tcp,local')


# Configuration for tlstester.py:
#
# Hostname to connect to, must match exactly what tsltester.py is signing the
# certificates for.
test_tls_tester_host = environ.get('TSTTLSTESTERHOST')
# Main port to connect to, tlstester.py will redirect the test cases to other
# ports as well.
test_tls_tester_port = environ.get('TSTTLSTESTERPORT')
# Set to true if tlstester.py's ca3.crt has been inserted into the system
# trusted root certificate store.
test_tls_tester_sys_store = environ.get('TSTTLSTESTERSYSSTORE', 'false').lower() == 'true'


_MONETDB_VERSION = None


def have_monetdb_version_at_least(major, minor, patch):
    global _MONETDB_VERSION
    if _MONETDB_VERSION is None:
        with pymonetdb.connect(**test_args) as conn, conn.cursor() as cursor:
            cursor.execute("SELECT value FROM environment WHERE name = 'monet_version'")
            _MONETDB_VERSION = tuple(int(part) for part in cursor.fetchone()[0].split('.'))

    return _MONETDB_VERSION >= (major, minor, patch)


try:
    import_module('lz4.frame')
    test_have_lz4 = True
except ModuleNotFoundError:
    test_have_lz4 = False


# Debug the debugging
if __name__ == "__main__":
    print(f'test_mapi_args = {test_mapi_args!r}')
    print(f'control_test_args = {control_test_args!r}')
    print(f'test_args = {test_args!r}')
    print(f'test_url = {test_url!r}')
    print(f'test_full = {test_full!r}')
    print(f'test_control = {test_control!r}')
    print(f'test_tls_tester_host = {test_tls_tester_host!r}')
    print(f'test_tls_tester_port = {test_tls_tester_port!r}')
    print(f'test_tls_tester_sys_store = {test_tls_tester_sys_store!r}')
    print(f'test_have_lz4 = {test_have_lz4!r}')
    try:
        print()
        have_monetdb_version_at_least(0, 0, 0)
        print(f'_MONETDB_VERSION = {_MONETDB_VERSION!r}')
    except Exception as e:
        print(f'Error while retrieving MonetDB version number: {e}')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

from unittest import TestCase
import pymonetdb
from tests.util import test_args


class ExecuteTests(TestCase):
    def test_execute_return_value(self):
        # The return value of Cursor.execute() is not specified by PEP 249
        # but we want it to behave as follows:

        conn = pymonetdb.connect(**test_args)
        c = conn.cursor()

        ret = c.execute("DROP TABLE IF EXISTS foo")
        self.assertIsNone(ret)

        ret = c.execute("CREATE TABLE foo(i INT)")
        self.assertIsNone(ret)

        ret = c.execute("INSERT INTO foo SELECT * FROM sys.generate_series(0,10)")
        self.assertEqual(ret, 10)

        ret = c.execute("DELETE FROM foo WHERE i % 3 = 0")
        self.assertEqual(ret, 4)  # 0 3 6 9

        ret = c.execute("SELECT * FROM foo")
        self.assertEqual(ret, 6)  # 1 2 4 5 7 8

        ret = c.execute("DROP TABLE foo")
        self.assertIsNone(ret)

        conn.rollback()



"""
Utilities for parsing MonetDB URLs
"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.


import re
from typing import Any, Callable, Union
from urllib.parse import parse_qsl, urlparse, quote as urlquote


def looks_like_url(text: str) -> bool:
    return (
        text.startswith("mapi:")
        or text.startswith("monetdb:")
        or text.startswith("monetdbs:")
        or text.startswith("monetdbe:")
    )


# Note that 'valid' is not in VIRTUAL:
CORE = set(['tls', 'host', 'port', 'database', 'tableschema', 'table'])
KNOWN = set([
    'tls', 'host', 'port', 'database', 'tableschema', 'table',
    'sock', 'sockdir', 'sockprefix', 'cert', 'certhash', 'clientkey', 'clientcert',
    'user', 'password', 'language', 'autocommit', 'schema', 'timezone',
    'binary', 'replysize', 'fetchsize', 'maxprefetch',
    'connect_timeout']
)
IGNORED = set(['hash', 'debug', 'logfile'])
VIRTUAL = set([
    'connect_scan', 'connect_unix', 'connect_tcp', 'connect_port',
    'connect_tls_verify', 'connect_certhash_digits',
    'connect_binary', 'connect_clientkey', 'connect_clientcert',
])

_BOOLEANS = dict(
    true=True,
    false=False,
    yes=True,
    no=False,
    on=True,
    off=False
)

_DEFAULTS = dict(
    tls=False,
    host="",
    port=-1,
    database="",
    tableschema="",
    table="",
    sock="",
    sockdir="/tmp",
    sockprefix=".s.monetdb.",
    cert="",
    certhash="",
    clientkey="",
    clientcert="",
    user="monetdb",
    password="monetdb",
    language="sql",
    autocommit=False,
    schema="",
    timezone=None,
    binary="on",
    replysize=None,
    fetchsize=None,
    maxprefetch=None,
    connect_timeout=0,
    dangerous_tls_nocheck="",
)


def parse_bool(x: Union[str, bool]):
    if isinstance(x, bool):
        return x
    try:
        return _BOOLEANS[x.lower()]
    except KeyError:
        raise ValueError("invalid boolean value")


class urlparam:
    """Decorator to create getter/setter for url parameter on a Target instance"""

    field: str
    parser: Callable[[Union[str, Any]], Any]

    def __init__(self, name, typ, doc):
        self.field = name
        if typ == 'string' or typ == 'path':
            self.parser = str
        elif typ == 'integer':
            self.parser = int
        elif typ == 'bool':
            self.parser = parse_bool
        else:
            raise ValueError(f"invalid type '{typ}'")
        self.__doc__ = doc

    def __get__(self, instance, owner):
        # don't know the meaning of the owner parameter; irrelevant?
        return instance._VALUES.get(self.field)

    def __set__(self, instance, value):
        parsed = (self.parser)(value)
        instance._VALUES[self.field] = parsed
        if self.field in instance._TOUCHED:
            instance._TOUCHED[self.field] = True

    def __delete__(self, instance):
        raise Exception("cannot delete url parameter")


class Target:
    """Holds all parameters needed to connect to MonetDB."""
    __slots__ = [
        '_VALUES',
        '_OTHERS',
        '_TOUCHED',
    ]

    def __init__(self, *, prototype=None):
        if prototype:
            self._VALUES = {**prototype._VALUES}
            self._OTHERS = {**prototype._OTHERS}
            self._TOUCHED = {**prototype._TOUCHED}
        else:
            self._VALUES = dict(**_DEFAULTS)
            self._OTHERS = {}
            self._TOUCHED = dict(user=False, password=False)

    def clone(self):
        return Target(prototype=self)

    tls = urlparam('tls', 'bool', 'secure the connection using TLS')
    host = urlparam(
        'host', 'string', 'IP number, domain name or one of the special values `localhost` and `localhost.`')
    port = urlparam('port', 'integer',
                    'TCP port, also used to pick Unix Domain socket path')
    database = urlparam('database', 'string', 'name of database to connect to')
    tableschema = urlparam('tableschema', 'string', 'only used for REMOTE TABLE, otherwise unused')
    table = urlparam('table', 'string', 'only used for REMOTE TABLE, otherwise unused')
    sock = urlparam('sock', 'path', 'path to Unix Domain socket to connect to')
    sockdir = urlparam('sockdir', 'path', 'directory where implicit Unix domain sockets are created')
    sockprefix = urlparam('sockprefix', 'string', 'prefix for implicit Unix domain sockets')
    cert = urlparam(
        'cert', 'path', 'path to TLS certificate to authenticate server with')
    certhash = urlparam(
        'certhash', 'string', 'hash of server TLS certificate must start with these hex digits; overrides cert')
    clientkey = urlparam(
        'clientkey', 'path', 'path to TLS key (+certs) to authenticate with as client')
    clientcert = urlparam(
        'clientcert', 'path', "path to TLS certs for 'clientkey', if not included there")
    user = urlparam('user', 'string', 'user name to authenticate as')
    password = urlparam('password', 'string', 'password to authenticate with')
    language = urlparam('language', 'string',
                        'for example, "sql", "mal", "msql", "profiler"')
    autocommit = urlparam('autocommit', 'bool', 'initial value of autocommit')
    schema = urlparam('schema', 'string', 'initial schema')
    timezone = urlparam('timezone', 'integer',
                        'client time zone as minutes east of UTC')
    binary = urlparam(
        'binary', 'string', 'whether to use binary result set format (number or bool)')
    replysize = urlparam('replysize', 'integer',
                         'rows beyond this limit are retrieved on demand, <1 means unlimited')
    maxprefetch = urlparam('maxprefetch', 'integer', 'specific to pymonetdb')
    connect_timeout = urlparam('connect_timeout', 'integer', 'abort if connect takes longer than this')
    dangerous_tls_nocheck = urlparam(
        'dangerous_tls_nocheck', 'bool',
        'comma separated certificate checks to skip, host: do not verify host, cert: do not verify certificate chain')

    # alias
    fetchsize = replysize

    def set(self, key: str, value: str):
        if key in KNOWN:
            setattr(self, key, value)
        elif key in IGNORED or '_' in key:
            self._OTHERS[key] = value
        else:
            raise ValueError(f"unknown parameter {key!r}")

    def get(self, key: str):
        if key in KNOWN or key in VIRTUAL:
            return getattr(self, key)
        elif key in IGNORED or '_' in key:
            return self._OTHERS[key]
        else:
            raise KeyError(key)

    def boundary(self):
        """If user was set and password wasn't, clear password"""
        if self._TOUCHED['user'] and not self._TOUCHED['password']:
            self.password = ''
        self._TOUCHED['user'] = False
        self._TOUCHED['password'] = False

    def summary_url(self):
        db = self.database or ''
        if self.sock:
            return f"monetdb://localhost/{db}?sock={urlquote(self.sock)}"
        scheme = "monetdbs" if self.tls else "monetdb"
        host = self.host or "localhost"
        if self.port and self.port > 0 and self.port != 50_000:
            return f"{scheme}://{host}:{self.port}/{db}"
        else:
            return f"{scheme}://{host}/{db}"

    def parse(self, url: str):
        self.boundary()
        if url.startswith("monetdb://") or url.startswith("monetdbs://"):
            self._set_core_defaults()
            self._parse_monetdb_url(url)
        elif url.startswith("mapi:monetdb://"):
            self._set_core_defaults()
            self._parse_mapi_monetdb_url(url)
        else:
            raise ValueError("URL must start with monetdb://, monetdbs:// or mapi:monetdb://")
        self.boundary()

    def _set_core_defaults(self):
        self.tls = False
        self.host = ''
        self.port = _DEFAULTS['port']
        self.database = ''

    def _parse_monetdb_url(self, url):    # noqa C901
        parsed = urlparse(url, allow_fragments=True)

        if parsed.scheme == 'monetdb':
            self.tls = False
        elif parsed.scheme == 'monetdbs':
            self.tls = True
        else:
            raise ValueError(f"Invalid URL scheme: {parsed.scheme}")

        if parsed.hostname is not None:
            host = strict_percent_decode('host name', parsed.hostname)
            if host == 'localhost':
                host = ''
            elif host == 'localhost.':
                host = 'localhost'
            self.host = host
        if parsed.port is not None:
            port = parsed.port
            if port is not None and not 1 <= port <= 65535:
                raise ValueError(f"Invalid port number: {port}")
            self.port = port

        path = parsed.path
        if path:
            parts = path.split("/")
            # 0: before leading slash, always empty
            # 1: database name
            # 2: schema name, ignored
            # 3: table name, ignored
            # more: error
            assert parts[0] == ""
            if len(parts) > 4:
                raise ValueError("invalid table name: " + '/'.join(parts[3:]))
            self.database = strict_percent_decode('database name', parts[1])
            if len(parts) > 2:
                self.tableschema = strict_percent_decode('schema name', parts[2])
            if len(parts) > 3:
                self.table = strict_percent_decode('table name', parts[3])

        for key, value in parse_qsl(parsed.query, keep_blank_values=True, strict_parsing=True):
            if not key:
                raise ValueError("empty key is not allowed")
            key = strict_percent_decode(repr(key), key)
            value = strict_percent_decode(f"value of {key!r}", value)
            if key in CORE:
                raise ValueError(
                    "key {key!r} is not allowed in the query parameters")
            self.set(key, value)

    def _parse_mapi_monetdb_url(self, url):    # noqa C901
        # mapi urls have no percent encoding at all
        parsed = urlparse(url[5:])
        if parsed.scheme != 'monetdb':
            raise ValueError(f"Invalid scheme {parsed.scheme!r}")
        self.tls = False
        if parsed.username is not None:
            self.user = parsed.username
        if parsed.password is not None:
            self.password = parsed.password
        if parsed.hostname is not None:
            self.host = parsed.hostname
        if parsed.port is not None:
            self.port = parsed.port

        path = parsed.path
        if path is not None:
            if parsed.hostname is None and parsed.port is None:
                self.sock = path
            else:
                path = path[1:]
                self.database = path  # validation will happen later

        # parse query manually, the library functions perform percent decoding
        if not parsed.query:
            return
        for part in parsed.query.split('&'):
            # language
            if part.startswith('language='):
                self.language = part[9:]
            elif part.startswith('database='):
                self.database = part[9:]
            elif part.startswith('user=') or part.startswith('password='):
                # ignored because libmapi does so
                pass
            elif part.startswith('binary='):
                # pymonetdb-only, backward compat
                self.binary = part[7:]
            elif part.startswith('replysize='):
                # pymonetdb-only, backward compat
                self.replysize = part[10:]
            elif part.startswith('maxprefetch='):
                # pymonetdb-only, backward compat
                self.maxprefetch = part[12:]
            else:
                # unknown parameters are ignored
                pass

    def _parse_mapi_merovingian_url(self, url):    # noqa C901
        # mapi urls have no percent encoding at all
        parsed = urlparse(url[5:])
        if parsed.scheme != 'merovingian':
            raise ValueError(f"Invalid scheme {parsed.scheme!r}")
        if parsed.username is not None:
            self.user = parsed.username
        if parsed.password is not None:
            self.password = parsed.password
        if parsed.hostname is not None:
            self.host = parsed.hostname
        if parsed.port is not None:
            self.port = parsed.port

        path = parsed.path
        if path is not None:
            if parsed.hostname is None and parsed.port is None:
                self.sock = path
            else:
                path = path[1:]
                self.database = path  # validation will happen later

        # parse query manually, the library functions perform percent decoding
        if not parsed.query:
            return
        for part in parsed.query.split('&'):
            # language
            if part.startswith('language='):
                self.language = part[9:]
            elif part.startswith('database='):
                self.database = part[9:]
            elif part.startswith('user=') or part.startswith('password='):
                # ignored because libmapi does so
                pass
            elif part.startswith('binary='):
                # pymonetdb-only, backward compat
                self.binary = part[7:]
            elif part.startswith('replysize='):
                # pymonetdb-only, backward compat
                self.replysize = part[10:]
            elif part.startswith('maxprefetch='):
                # pymonetdb-only, backward compat
                self.maxprefetch = part[12:]
            else:
                # unknown parameters are ignored
                pass

    def validate(self):    # noqa C901
        # 1. The parameters have the types listed in the table in [Section
        #    Parameters](#parameters).
        #
        # This has already been checked by the url_param magic.

        # 2. At least one of **sock** and **host** must be empty.
        if self.sock and self.host:
            raise ValueError("With sock=, host must be empty or 'localhost'")

        # 3. The string parameter **binary** must either parse as a boolean or as a
        #    non-negative integer.
        #
        # Let connect_binary do all the work.
        if self.connect_binary(1) < 0:
            raise ValueError("Parameter 'binary' must be ≥ 0")

        # 4. If **sock** is not empty, **tls** must be 'off'.
        if self.sock and self.tls:
            raise ValueError("TLS cannot be used with Unix domain sockets")

        # 5. If **certhash** is not empty, it must be of the form `{sha256}hexdigits`
        #    where hexdigits is a non-empty sequence of 0-9, a-f, A-F and colons.
        if self.certhash and not _HASH_PATTERN.match(self.certhash):
            raise ValueError(f"invalid certhash: {self.certhash}")

        # 6. If **tls** is 'off', **cert** and **certhash** must be 'off' as well.
        if not self.tls and (self.cert or self.certhash):
            raise ValueError("'cert' and 'certhash' can only be used with monetdbs:")

        # 7. Parameters **database**, **tableschema** and **table** must consist only of
        #    upper- and lowercase letters, digits, dashes and underscores. They must not
        #    start with a dash.
        if self.database and not _DATABASE_PATTERN.match(self.database):
            raise ValueError(f"invalid database name {self.database!r}")
        if self.tableschema and not _DATABASE_PATTERN.match(self.tableschema):
            raise ValueError(f"invalid schema name {self.tableschema!r}")
        if self.table and not _DATABASE_PATTERN.match(self.table):
            raise ValueError(f"invalid table name {self.table!r}")

        # 8. Parameter **port**, if present, must be in the range 1-65535.
        if self.port != -1 and not 1 <= self.port <= 65535:
            raise ValueError(f"Invalid port number: {self.port}")

        # 9. If **clientcert** is set, **clientkey** must also be set.
        if self.clientcert and not self.clientkey:
            raise ValueError("clientcert can only be used together with clientkey")

    @property
    def connect_scan(self):
        if not self.database:
            return False
        if self.sock or self.host or self.port != -1:
            return False
        if self.tls:
            return False
        return True

    @property
    def connect_unix(self):
        if self.sock:
            return self.sock
        if self.tls:
            return ""
        if self.host == "":
            return f"{self.sockdir}/{self.sockprefix}{self.connect_port}"
        return ""

    @property
    def connect_tcp(self):
        if self.sock:
            return ""
        return self.host or "localhost"

    @property
    def connect_port(self):
        assert self.port == -1 or 1 <= self.port <= 65535
        if self.port == -1:
            return 50000
        else:
            return self.port

    @property
    def connect_tls_verify(self):
        if not self.tls:
            return ""
        if self.certhash:
            return "hash"
        if self.cert:
            return "cert"
        return "system"

    @property
    def connect_clientkey(self):
        return self.clientkey

    @property
    def connect_clientcert(self):
        return self.clientcert or self.clientkey

    def connect_binary(self, max: int):
        try:
            return int(self.binary)
        except ValueError:
            try:
                return max if parse_bool(self.binary) else 0
            except ValueError:
                raise ValueError("invalid value for 'binary': {self.binary}, must be int or bool")

    @property
    def connect_certhash_digits(self):
        m = _HASH_PATTERN.match(self.certhash)
        if m:
            return m.group(1).lower().replace(':', '')
        else:
            return None


_UNQUOTE_PATTERN = re.compile(b"[%](.?.?)")
_DATABASE_PATTERN = re.compile("^[A-Za-z0-9_][-A-Za-z0-9_.]*$")
_HASH_PATTERN = re.compile(r"^sha256:([0-9a-fA-F:]+)$")


def _unquote_fun(m) -> bytes:
    digits = m.group(1)
    if len(digits) != 2:
        raise ValueError()
    return bytes([int(digits, 16)])


def strict_percent_decode(context: str, text: str) -> str:
    try:
        return str(_UNQUOTE_PATTERN.sub(_unquote_fun, bytes(text, "ascii")), "utf-8")
    except (ValueError, UnicodeDecodeError) as e:
        raise ValueError("invalid percent escape in {context}") from e



# coding: utf-8

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

from abc import abstractmethod
import datetime
from decimal import ROUND_HALF_UP, Decimal
from random import Random
from typing import Any, Callable, List, Optional, Tuple
from unittest import SkipTest, TestCase
from uuid import UUID
import pymonetdb
from tests.util import have_monetdb_version_at_least, test_args

QUERY_TEMPLATE = """\
WITH resultset AS (
    SELECT
        %(exprs)s,
        42 AS dummy
    FROM sys.generate_series(0, %(count)s - 1)
),
t AS (SELECT 42 AS dummy UNION SELECT 43)
--
SELECT * FROM
    resultset RIGHT OUTER JOIN t
    ON resultset.dummy = t.dummy;
"""


def decimal_column(p, s):
    dec = f"DECIMAL({p}, {s})"
    expr = f"CAST(CAST(value AS {dec}) * 1.5 AS {dec})"
    # MonetDB rounds (N + 0.5) away from zero
    quantum = Decimal('10') ** (-s)

    def verifier(n):
        return (Decimal(n) * 3 / 2).quantize(quantum, rounding=ROUND_HALF_UP)

    return (expr, verifier)


test_uuid = UUID('{12345678-1234-5678-1234-567812345678}')
test_blobs = {
    0: b'MONETDB',
    1: b'',
    2: None,
}


def seconds_timedelta_helper(value, multiplier):
    t = Decimal(multiplier) * value
    millis = int(1000 * t)
    return datetime.timedelta(milliseconds=millis)


TEST_COLUMNS = dict(
    int_col=("CAST(value AS int)", lambda n: n),
    tinyint_col=("CAST(value % 128 AS tinyint)", lambda n: n % 128),
    smallint_col=("CAST(value AS smallint)", lambda n: n),
    bigint_col=("CAST(value AS bigint)", lambda n: n),
    # hugeint_col=("CAST(value AS hugeint)", lambda n: n),    text_col=("'v' || value", lambda n: f"v{n}"),
    text_col=("'v' || value", lambda n: f"v{n}"),
    bool_col=("(value % 2 = 0)", lambda n: (n % 2) == 0),
    decimal_col=decimal_column(5, 2),
    real_col=("CAST(value AS REAL) / 2", lambda x: x / 2),
    float_col=("CAST(value AS FLOAT) / 2", lambda x: x / 2),
    double_col=("CAST(value AS DOUBLE) / 2", lambda x: x / 2),
    f32_col=("CAST(value AS float(24)) / 2", lambda x: x / 2),
    f53_col=("CAST(value AS float(53)) / 2", lambda x: x / 2),
    blob_col=(
        "CAST((CASE WHEN value % 3 = 0 THEN '4d4f4e45544442' WHEN value % 3 = 1 THEN '' ELSE NULL END) AS BLOB)",
        lambda x: test_blobs[x % 3]),
    months_col=("CAST(CAST(value AS TEXT) AS INTERVAL MONTH)", lambda x: x),
    days_col=("CAST(CAST(value AS TEXT) AS INTERVAL DAY) * 1.007", lambda x: int(x * Decimal('1.007'))),
    seconds_col=("CAST(CAST(value AS TEXT) AS INTERVAL SECOND) * 1.007", lambda x: seconds_timedelta_helper(x, '1.007')),
    # not a very dynamic example:
    uuid_col=("CAST('12345678-1234-5678-1234-567812345678' AS UUID)", lambda x: test_uuid)
)

# Some versions of MonetDB have a bug where the output is wrong if more than one
# of the following is present in the result set at the same time.
# We'll test them separately
BLACKLIST = set(['months_col', 'days_col', 'seconds_col'])


class BaseTestCases(TestCase):
    _server_binexport_level: Optional[int] = None
    _server_has_huge: Optional[bool] = None
    conn: Optional[pymonetdb.Connection] = None
    cursor: Optional[pymonetdb.sql.cursors.Cursor] = None
    cur: int = 0
    rowcount: int = 0
    verifiers: List[Tuple[str, Callable[[int], Any]]] = []
    colnames: List[str] = []

    def probe_server(self):
        if self._server_binexport_level is not None:
            return
        conn = self.connect_with_args()
        self._server_binexport_level = conn.mapi.binexport_level
        cursor = conn.cursor()
        cursor.execute("SELECT sqlname FROM sys.types WHERE sqlname = 'hugeint'")
        self._server_has_huge = cursor.rowcount > 0
        cursor.close()
        conn.close()

    def have_binary(self, at_least=1):
        self.probe_server()
        return self._server_binexport_level >= at_least

    def server_has_new_time_conversion(self):
        return have_monetdb_version_at_least(11, 50, 0)

    def have_huge(self):
        self.probe_server()
        return self._server_has_huge

    def skip_unless_have_binary(self):
        if not self.have_binary():
            raise SkipTest("need server with support for binary")

    def skip_unless_have_huge(self):
        if not self.have_huge():
            raise SkipTest("need server with support for hugeint")

    def setUp(self):
        self.cur = 0
        self.rowcount = 0
        self.to_close = []

    def close_connection(self):
        if self.cursor:
            self.cursor.close()
            self.cursor = None

    def tearDown(self):
        if self.cursor:
            self.cursor.execute("ROLLBACK")
        for c in self.to_close:
            try:
                c.close()
            except Exception:
                pass

    def connect_with_args(self, **kw_args) -> pymonetdb.Connection:
        try:
            args = dict()
            args.update(test_args)
            args.update(kw_args)
            conn = pymonetdb.connect(**args)   # type: ignore
        except AttributeError:
            self.fail("No connect method found in pymonetdb module")
        self.to_close.append(conn)
        with conn.cursor() as c:
            c.execute("SELECT %s", f"This connection is for test {self.id()!r}")
        return conn

    @abstractmethod
    def setup_connection(self) -> Tuple[pymonetdb.Connection, Optional[int]]:
        assert False

    def assertAtEnd(self):
        self.assertEqual(self.rowcount, self.cur, f"expected to be at end ({self.rowcount} rows), not {self.cur}")

    def do_connect(self):
        if self.conn is None:
            assert self.cursor is None
            self.conn, self.expect_binary_after = self.setup_connection()
            self.cursor = self.conn.cursor()

    def construct_query(self, n, cols):
        if isinstance(cols, dict):
            test_columns = cols
        else:
            test_columns = dict()
            for col in cols:
                test_columns[col] = TEST_COLUMNS[col]

        exprs = []
        verifiers = []
        colnames = []
        for col, (expr, verifier) in test_columns.items():
            exprs.append(f"{expr} AS {col}")
            verifiers.append(verifier)
            colnames.append(col)
        query = QUERY_TEMPLATE % dict(
            exprs=",\n        ".join(exprs),
            count=n
        )

        return query, colnames, verifiers

    def set_expected_results(self, rowcount, colnames, verifiers):
        self.cur = 0
        self.rowcount = rowcount
        self.colnames = colnames
        self.verifiers = verifiers

    def do_query(self, n, cols=('int_col',)):
        query, colnames, verifiers = self.construct_query(n, cols)

        self.do_connect()
        self.cursor.execute(query)
        self.set_expected_results(n, colnames, verifiers)

        self.assertEqual(n, self.cursor.rowcount)

    def verifyRow(self, n, row):
        # two dummy columns because of the outer join
        self.assertEqual(len(row) - 2, len(self.verifiers))

        if n == self.rowcount - 1:
            expected = len(self.verifiers) * (None,)
        else:
            expected = tuple(verifier(n) for verifier in self.verifiers)

        found = row[:len(self.verifiers)]
        if found == expected:
            return

        for i in range(len(self.verifiers)):
            if found[i] == expected[i]:
                continue
            self.assertEqual(expected[i], found[i], f"Mismatch at row {n}, col {i} '{self.colnames[i]}'")

        self.assertEqual(expected, found, f"Mismatch at row {n}")

    def verifyBinary(self):
        if not self.have_binary():
            return
        if self.expect_binary_after is None:
            return
        if self.cur <= self.expect_binary_after:
            return
        # with this many rows, binary should have been used
        self.assertTrue(self.cursor.used_binary_protocol(), "Expected binary result sets to be used")

    def do_fetchone(self):
        row = self.cursor.fetchone()
        if self.cur < self.rowcount:
            self.assertIsNotNone(row)
            self.verifyRow(self.cur, row)
            self.cur += 1
        else:
            self.assertIsNone(row)
        self.verifyBinary()

    def do_fetchmany(self, n):
        rows = self.cursor.fetchmany(n)
        if n is not None:
            expectedRows = min(n, self.rowcount - self.cur)
            self.assertEqual(expectedRows, len(rows))
        for i, row in enumerate(rows):
            self.verifyRow(self.cur + i, row)
        self.cur += len(rows)
        self.verifyBinary()

    def do_fetchall(self) -> int:
        oldcur = self.cur
        assert self.cursor
        rows = self.cursor.fetchall()
        expectedRows = self.rowcount - self.cur
        self.assertEqual(expectedRows, len(rows))
        for i, row in enumerate(rows):
            self.verifyRow(self.cur + i, row)
        self.cur += len(rows)
        self.verifyBinary()
        self.assertAtEnd()
        return self.cur - oldcur

    def do_scroll(self, n, mode):
        self.cursor.scroll(n, mode)
        if mode == 'absolute':
            self.cur = n
        elif mode == 'relative':
            self.cur += n
        else:
            raise self.failureException(f"unexpected mode {mode}")

    def test_fetchone(self, n=1000):
        self.do_query(n)
        for i in range(self.rowcount + 1):
            self.do_fetchone()
        self.assertAtEnd()

    def test_fetchmany(self, n=1000):
        self.do_query(n)
        for i in range(n):
            self.do_fetchmany(None)
        self.assertAtEnd()

    def test_fetchmany42(self, n=1000):
        self.do_query(n)
        for i in range(self.rowcount // 42 + 1):
            self.do_fetchmany(42)
        self.assertAtEnd()

    def test_fetchmany120(self, n=1000):
        self.do_query(n)
        for i in range(self.rowcount // 120 + 1):
            self.do_fetchmany(120)
        self.assertAtEnd()

    def test_fetchall(self, n=1000):
        self.do_query(n)
        self.do_fetchall()
        self.assertAtEnd()

    def test_fetchone_large(self):
        self.test_fetchone(25_000)

    def test_fetchmany_large(self):
        self.test_fetchmany(100_000)

    def test_fetchmany42_large(self):
        self.test_fetchmany42(100_000)

    def test_fetchmany120_large(self):
        self.test_fetchmany120(100_000)

    def test_fetchall_large(self):
        self.test_fetchall(100_000)

    def test_scroll(self):
        rng = Random()
        rng.seed(42)
        self.do_query(1000)
        for _ in range(500):
            x = rng.randrange(0, self.rowcount)
            y = rng.randrange(0, self.rowcount)
            if x > y:
                (x, y) = (y, x)
            if rng.randrange(0, 10) >= 2:
                y = rng.randrange(x, min(y, self.rowcount))
            if rng.randrange(0, 2) > 0:
                self.do_scroll(x, 'absolute')
            else:
                self.do_scroll(x - self.cur, 'relative')
            self.do_fetchmany(y - x)

    def test_multiple(self):
        self.do_connect()

        nrows = 20
        query1, colnames1, verifiers1 = self.construct_query(nrows, ['int_col'])
        query2, colnames2, verifiers2 = self.construct_query(2 * nrows, ['tinyint_col', 'text_col'])

        def access_pattern(self, n):
            self.do_fetchone()
            self.do_fetchmany(n // 2)
            self.do_scroll(n // 4, 'absolute')
            remaining = self.do_fetchall()
            self.assertEqual(remaining, n // 4 * 3)
            self.verifyBinary()
            self.assertAtEnd()

        query = f"{query1};\n{query2}"
        self.cursor.execute(query)

        self.set_expected_results(nrows, colnames1, verifiers1)
        access_pattern(self, nrows)

        self.assertTrue(self.cursor.nextset())
        self.set_expected_results(2 * nrows, colnames2, verifiers2)
        access_pattern(self, 2 * nrows)

        self.assertFalse(self.cursor.nextset())

    def test_huge(self):
        self.skip_unless_have_huge()
        max_value = (1 << 127) - 1
        min_value = - max_value
        columns = dict(
            up_from_0=("CAST(value AS hugeint)", lambda n: n),
            down_from_0=("-CAST(value AS hugeint)", lambda n: -n),
            up_from_min=(f"{min_value} + CAST(value AS hugeint)", lambda n: min_value + n),
            down_from_max=(f"{max_value} - CAST(value AS hugeint)", lambda n: max_value - n),
        )
        self.do_query(5, columns)
        self.do_fetchall()
        self.verifyBinary()

    def test_data_types(self):
        self.do_query(250, TEST_COLUMNS.keys() - BLACKLIST)
        self.do_fetchall()
        # no self.verifyBinary()

    def test_binary_data_types(self):
        self.skip_unless_have_binary()
        self.do_query(250, TEST_COLUMNS.keys() - BLACKLIST)
        self.do_fetchall()
        self.verifyBinary()

    def test_decimal_types(self):
        cases = set()
        widest = 39 if self.have_huge() else 18
        for p in range(1, widest):
            cases.add((p, 0))
            cases.add((p, min(3, p - 1)))

        test_columns = dict()
        for p, s in sorted(cases):
            name = f"dec_{p}_{s}"
            maker = decimal_column(p, s)
            test_columns[name] = maker

        # only get 6 rows because 1.5 * n must fit in DECIMAL(1,0)
        self.do_query(6, test_columns)
        self.do_fetchall()
        self.verifyBinary()

    def prepare_temporal_tests(self, col):
        self.do_connect()
        minutes_east = 60 + 30  # easily recognizable
        self.conn.set_timezone(60 * minutes_east)
        self.conn.set_autocommit(False)

        our_timezone = datetime.timezone(datetime.timedelta(hours=1, minutes=30))

        self.cursor.execute("DROP TABLE IF EXISTS foo")
        cols = [
            "name TEXT",
            "ts_with TIMESTAMPTZ",
            "ts_without TIMESTAMP",
            "d DATE",
            "t_with TIMETZ",
            "t_without TIME",
        ]
        create_statement = f"CREATE TABLE foo ({(', '.join(cols))})"
        self.cursor.execute(create_statement)
        interesting_times = [
            ('dummy', "NULL"),

            ('null', "NULL"),
            # MonetDB will normalize the following two into the exact same thing
            ('apollo13_utc', "TIMESTAMPTZ '1970-04-17 18:07:41+00:00'"),
            ('apollo13_pacific', "TIMESTAMPTZ '1970-04-17 10:07:41-08:00'"),
        ]
        insert_statement = (
            "INSERT INTO foo(name, ts_with) VALUES "
            + ", ".join(f"('{name}', {expr})" for name, expr in interesting_times)
        )
        self.cursor.execute(insert_statement)
        self.cursor.execute("UPDATE foo set ts_without = ts_with, d = ts_with, t_with = ts_with, t_without = ts_with")

        self.cursor.execute(f"SELECT name, {col} FROM foo")
        rows = self.cursor.fetchall()

        # needed by verifyBinary
        # update cur manually because it is set by do_fetchall but not
        # by cursor.fetchall
        self.cur = self.cursor.rowcount

        values = dict()
        for row in rows:
            name = row[0]
            # make REALLY REALLY sure the indices match the order in the SELECT clause!
            values[name] = row[1]

        self.verifyBinary()
        return (our_timezone, values)

    def test_temporal_timestamp_with_timezone(self):
        (tz, values) = self.prepare_temporal_tests('ts_with')
        self.assertIsNone(values['null'])

        # apollo13_utc was given as 1970-04-17 18:07:41+00:00,
        # stored in ts_with as 1970-04-17 18:07:41 UTC,
        # rendered on the wire as 1970-04-17 19:37+01:30,
        # converted to a DateTime of 19:37 in the +01:30 time zone
        x = values['apollo13_utc']
        self.assertEqual('1970-04-17T19:37:41+01:30', x.isoformat())
        self.assertEqual(tz, x.tzinfo)

        # apollo13_pacific was given as 1970-04-17 10:07:41-08:00,
        # stored in ts_with as 1970-04-17 18:07:41 UTC
        # rendered on the wire as 1970-04-17 19:37+01:30
        # converted to a DateTime of 19:37 in the +01:30 time zone
        x = values['apollo13_pacific']
        self.assertEqual('1970-04-17T19:37:41+01:30', x.isoformat())
        self.assertEqual(tz, x.tzinfo)

    def test_temporal_timestamp_without_timezone_old(self):
        # the difference between _old and _new is marked with an arrow <---
        if self.server_has_new_time_conversion():
            raise SkipTest("test only applies to old MonetDB versions")

        (tz, values) = self.prepare_temporal_tests('ts_without')
        self.assertIsNone(values['null'])

        # apollo13_utc was given as 1970-04-17 18:07:41+00:00,
        # stored in ts_with as 1970-04-17 18:07:41 UTC,
        # then stored in ts_without as 1970-04-17 18:07:41,   <---
        # rendered on the wire as 1970-04-17 18:07:41,
        # converted to a DateTime of 18:07 without timezone
        x = values['apollo13_utc']
        self.assertEqual('1970-04-17T18:07:41', x.isoformat())
        self.assertIsNone(x.tzinfo)

        # apollo13_pacific was given as 1970-04-17 10:07:41-08:00,
        # stored in ts_with as 1970-04-17 18:07:41 UTC,
        # then stored in ts_without as 1970-04-17 18:07:41,   <---
        # rendered on the wire as 1970-04-17 18:07:41,
        # converted to a DateTime of 18:07 without timezone
        x = values['apollo13_pacific']
        self.assertEqual('1970-04-17T18:07:41', x.isoformat())
        self.assertIsNone(x.tzinfo)

    def test_temporal_timestamp_without_timezone_new(self):
        # the difference between _old and _new is marked with an arrow <---
        if not self.server_has_new_time_conversion():
            raise SkipTest("test only applies to new MonetDB versions")

        (tz, values) = self.prepare_temporal_tests('ts_without')
        self.assertIsNone(values['null'])

        # apollo13_utc was given as 1970-04-17 18:07:41+00:00,
        # stored in ts_with as 1970-04-17 18:07:41 UTC,
        # then stored in ts_without as 1970-04-17 19:37:41,   <---
        # rendered on the wire as 1970-04-17 19:37:41,
        # converted to a DateTime of 19:37 without timezone
        x = values['apollo13_utc']
        self.assertEqual('1970-04-17T19:37:41', x.isoformat())
        self.assertIsNone(x.tzinfo)

        # apollo13_pacific was given as 1970-04-17 10:07:41-08:00,
        # stored in ts_with as 1970-04-17 18:07:41 UTC,
        # then stored in ts_without as 1970-04-17 19:37:41,   <---
        # rendered on the wire as 1970-04-17 19:37:41,
        # converted to a DateTime of 19:37 without timezone
        x = values['apollo13_pacific']
        self.assertEqual('1970-04-17T19:37:41', x.isoformat())
        self.assertIsNone(x.tzinfo)

    def test_temporal_date(self):
        (tz, values) = self.prepare_temporal_tests('d')
        self.assertIsNone(values['null'])

        # apollo13_utc was given as 1970-04-17 18:07:41+00:00,
        # stored in ts_with as 1970-04-17 18:07:41 UTC,
        # then stored in d as 1970-04-17,
        # rendered on the wire as 1970-04-17,
        # converted to a Date of 1970-04-17
        x = values['apollo13_utc']
        self.assertEqual('1970-04-17', x.isoformat())

        # apollo13_pacific was given as 1970-04-17 10:07:41-08:00,
        # stored in ts_with as 1970-04-17 18:07:41 UTC,
        # then stored in d as 1970-04-17,
        # rendered on the wire as 1970-04-17,
        # converted to a Date of 1970-04-17
        x = values['apollo13_pacific']
        self.assertEqual('1970-04-17', x.isoformat())

    def test_temporal_time_with_timezone(self):
        (tz, values) = self.prepare_temporal_tests('t_with')
        self.assertIsNone(values['null'])

        # apollo13_utc was given as 1970-04-17 18:07:41+00:00,
        # stored in ts_with as 1970-04-17 18:07:41 UTC,
        # then stored in t_with as 18:07:41 UTC,
        # rendered on the wire as 19:37+01:30,
        # converted to a DateTime of 19:37 in the +01:30 time zone
        x = values['apollo13_utc']
        self.assertEqual('19:37:41+01:30', x.isoformat())
        self.assertEqual(tz, x.tzinfo)

        # apollo13_pacific was given as 1970-04-17 10:07:41-08:00,
        # stored in ts_with as 1970-04-17 18:07:41 UTC
        # then stored in t_with as 18:07:41 UTC,
        # rendered on the wire as 19:37+01:30,
        # converted to a DateTime of 19:37 in the +01:30 time zone
        x = values['apollo13_pacific']
        self.assertEqual('19:37:41+01:30', x.isoformat())
        self.assertEqual(tz, x.tzinfo)

    def test_temporal_time_without_timezone_old(self):
        # the difference between _old and _new is marked with an arrow <---
        if self.server_has_new_time_conversion():
            raise SkipTest("test only applies to old MonetDB versions")

        (tz, values) = self.prepare_temporal_tests('t_without')
        self.assertIsNone(values['null'])

        # apollo13_utc was given as 1970-04-17 18:07:41+00:00,
        # stored in ts_with as 1970-04-17 18:07:41 UTC,
        # then stored in t_without as 18:07:41,          <---
        # rendered on the wire as 18:07:41,
        # converted to a DateTime of 18:07 without timezone
        x = values['apollo13_utc']
        self.assertEqual('18:07:41', x.isoformat())
        self.assertIsNone(x.tzinfo)

        # apollo13_pacific was given as 1970-04-17 10:07:41-08:00,
        # stored in ts_with as 1970-04-17 18:07:41 UTC,
        # then stored in t_without as 18:07:41,          <---
        # rendered on the wire as 18:07:41,
        # converted to a DateTime of 18:07 without timezone
        x = values['apollo13_pacific']
        self.assertEqual('18:07:41', x.isoformat())
        self.assertIsNone(x.tzinfo)

    def test_temporal_time_without_timezone_new(self):
        # the difference between _old and _new is marked with an arrow <---
        if not self.server_has_new_time_conversion():
            raise SkipTest("test only applies to new MonetDB versions")

        (tz, values) = self.prepare_temporal_tests('t_without')
        self.assertIsNone(values['null'])

        # apollo13_utc was given as 1970-04-17 18:07:41+00:00,
        # stored in ts_with as 1970-04-17 18:07:41 UTC,
        # then stored in t_without as 19:37:41,          <---
        # rendered on the wire as 19:37:41,
        # converted to a DateTime of 19:37 without timezone
        x = values['apollo13_utc']
        self.assertEqual('19:37:41', x.isoformat())
        self.assertIsNone(x.tzinfo)

        # apollo13_pacific was given as 1970-04-17 10:07:41-08:00,
        # stored in ts_with as 1970-04-17 18:07:41 UTC,
        # then stored in t_without as 19:37:41,          <---
        # rendered on the wire as 19:37:41,
        # converted to a DateTime of 19:37 without timezone
        x = values['apollo13_pacific']
        self.assertEqual('19:37:41', x.isoformat())
        self.assertIsNone(x.tzinfo)

    def test_interval_second(self):
        self.do_query(250, ['seconds_col'])
        self.do_fetchall()
        self.verifyBinary()

    def test_interval_day(self):
        self.do_query(250, ['days_col'])
        self.do_fetchall()
        self.verifyBinary()

    def test_interval_month(self):
        self.do_query(250, ['months_col'])
        self.do_fetchall()
        self.verifyBinary()


class TestResultSet(BaseTestCases):
    def setup_connection(self):
        # no special connect parameters
        conn = self.connect_with_args()
        # if binary is enabled we expect to see it after row 100
        binary_after = 100
        return (conn, binary_after)


class TestResultSetNoBinary(BaseTestCases):
    def setup_connection(self):
        self.skip_unless_have_binary()  # test is not interesting if server does not support binary anyway
        conn = self.connect_with_args(binary=0)
        binary_after = None
        # we do not expect to see any binary
        return (conn, binary_after)


class TestResultSetForceBinary(BaseTestCases):
    def setup_connection(self):
        self.skip_unless_have_binary()
        # replysize 1 switches to binary protocol soonest, at the cost of more batches.
        conn = self.connect_with_args(binary=1, replysize=1)
        binary_after = 1
        return (conn, binary_after)


class TestResultSetFetchAllBinary(BaseTestCases):
    def setup_connection(self):
        self.skip_unless_have_binary()
        conn = self.connect_with_args(binary=1, replysize=-1)
        binary_after = 10
        return (conn, binary_after)


class TestResultSetFetchAllNoBinary(BaseTestCases):
    def setup_connection(self):
        conn = self.connect_with_args(binary=0, replysize=-1)
        binary_after = None
        return (conn, binary_after)


class TestResultSetNoPrefetch(BaseTestCases):
    def setup_connection(self):
        conn = self.connect_with_args(maxprefetch=0)
        binary_after = 100
        return (conn, binary_after)


# Make sure the abstract base class doesn't get executed
del BaseTestCases
